(********************************************************************)
(*                                                                  *)
(*                               OBJGFX0.PAS                        *)
(*                                                                  *)
(*                     Object Oriented Graphics Library             *)
(*                                                                  *)
(*                         Base Classes for ObjGfx*                 *)
(*                                                                  *)
(*                            Copyright (c) 1999                    *)
(*                      Mark J Iuzzolino, Daniel Robbins            *)
(*                                                                  *)
(*                       Portions Copyright (c) 1972                *)
(*                            Harold J Iuzzolino                    *)
(*                                                                  *)
(*                            All Rights Reserved                   *)
(*                                                                  *)
(*                                                                  *)
(*                                                                  *)
(********************************************************************)

(*
 *   Rev 1.0  1999/05/24 mji
 * Initial revision
 *)

(*
 * Presumed data types and sizes (for all you non-pascal programmers):
 *
 * boolean   : 8-bit boolean value
 * wordbool  : 16-bit boolean value
 * byte      : unsigned 8-bit value
 * shortint  : signed 8-bit value
 * char      : unsigned 8-bit value (character)
 * word      : unsigned 16-bit value
 * integer   : signed 16-bit value
 * longint   : signed 32-bit value
 * real      : 48-bit floating point
 * pointer   : 32-bit pointer (16-bit selector; 16-bit offset)
 * pointer48 : 48-bit pointer (16-bit selector; 32-bit offset)
 * string    : array of char; string length is string[0]
 *
 *)

{$A+,B-,D+,E+,F-,G+,I+,L+,N-,P-,Q-,R-,S+,T-,V+,X+,Y+}
{$M 16384,0}
unit ObjGfx0;

interface

const MaxScanLines = 4096;
const TransparentColour:longint = 0;
const ForceBanked:boolean = FALSE;
const ForceInt10h:boolean = FALSE;
const OldVidMode:word     = $3;
const SegC000:word        = $c000;
const Seg0000:word        = $0000;
const Seg0040:word        = $0040;
const SegA000:word        = $a000;
const SegF000:word        = $F000;

const RadToDeg            = 180/Pi;

type PLTEChunkEntry = record {PNG}
       R,G,B              : byte
     end;

type PalType = array[byte] of PLTEChunkEntry;
type RawPalType = array[0..767] of byte;

type Dword=record Loword,Hiword:word end;
type Pointer48  = Real;

type PLongArray = ^TLongArray;
     PWordArray = ^TWordArray;
     PByteArray = ^TByteArray;

     TLongArray = array [0..8191] of longint;
     TWordArray = array [0..16383] of Word;
     TByteArray = array [0..32767] of byte;

type pcxheader = Record
       manufacturer       : Byte;
       version            : Byte;
       encoding           : Byte;
       bits_per_pixel     : Byte;
       xmin, ymin         : Word;
       xmax, ymax         : Word;
       hdpi, vdpi         : Word;
       palette            : Array [0..47] of Byte;
       reserved           : Byte;
       colour_planes      : Byte;
       Bytes_per_line     : Word;
       palette_Type       : Word;
       HorzScreenSize     : Word;
       VertScreenSize     : Word;
       filler             : Array [0..53] of Byte;
     end;

type mooishiheader = Record
       ID_String          : array [0..7] of char;
       Escbytes           : array [0..3] of byte;
       hres               : word;
       vres               : word;
       reserved           : array [0..5] of byte;
     end;

type PNGSignature = record
       Sig1,Sig2          : longint;
     end;

type IHDRChunk = record
       width              : longint; {4 billion pixels wide!?}
       height             : longint;
       BitDepth           : byte;
       ColourType         : byte;
       Compression        : byte;
       Filter             : byte;
       Interlace          : byte;
     end;

type PNGChunk = record
       DataLength         : longint;
       DataType           : array[0..3] of char;
     end;

type GifHeader = record
       Signature          : array[0..2] of char;
       Version            : array[0..2] of char;
       ScreenWidth        : word;
       ScreenHeight       : word;
       Packd              : byte;
       BackgroundColour   : byte;
       AspectRatio        : byte;
     end;

type ilbmChunk = record
       ChunkID            : array[0..3] of char;
       size               : longint;
     end;

type BitMapHeaDer = record
       width,height       : word;
       left,top           : word;
       BitPlanes          : byte;
       Masking            : byte;
       Compress           : byte;
       Paddington         : byte;
       Transparency       : word;
       XAspectRatio       : byte;
       YAspectRatio       : byte;
       PageWidth          : word;
       PageHeight         : word;
     end;

type WinBMPFileHeader = record
       FileType           : word;
       FileSize           : longint;
       Reserved1          : word;   {always 0}
       Reserved2          : word;   {always 0}
       BitmapOffset       : longint;
     end;

type Win3xBitmapHeader = record
       FileHeader         : WinBMPFileHeader;
       Size               : longint;
       Width,Height       : longint;
       Planes             : word;
       BitsPerPixel       : word;
       { Fields added for Windows 3.x follow this line}
       Compression        : longint;
       SizeOfBitmap       : longint;
       HorzResolution     : longint;
       VertResolution     : longint;
       ColoursUsed        : longint;
       ColoursImportant   : longint;
     end;

type Win3xPaletteElement = record
       blue,green,red,pad : byte;
     end;

type TVESA_rec = record
       VBESignature       : array [0..3] of Char;
       minVersion         : Byte;
       majVersion         : Byte;
       OEMStringPtr       : Pchar;
       Capabilities       : LongInt;
       VideoModePtr       : ^TWordArray;
       TotalMemory        : word;

       {VESA 2.0}
       OemSoftwareRev     : word;
       OemVendorNamePtr   : Pchar;
       OemProductNamePtr  : Pchar;
       OemProductRevPtr   : Pchar;
       Paddington: array [35..512] of Byte;
     end;

type TMode_Rec = record
       ModeAttributes     : Word;
       WindowAFlags       : Byte;
       WindowBFlags       : Byte;
       Granularity        : Word;
       WindowSize         : Word;
       WindowASeg         : Word;
       WindowBSeg         : Word;
       BankSwitch         : Pointer;
       BytesPerLine       : Word;
       XRes,YRes          : Word;
       CharWidth          : Byte;
       CharHeight         : Byte;
       NumBitplanes       : Byte;
       BitsPerPixel       : Byte;
       NumberOfBanks      : Byte;
       MemoryModel        : Byte;
       BankSize           : Byte;
       NumOfImagePages    : byte;
       Reserved           : byte;
       {Direct Colour fields (required for Direct/6 and YUV/7 memory models}
       RedMaskSize        : byte;
       RedFieldPosition   : Byte;
       GreenMaskSize      : Byte;
       GreenFieldPosition : Byte;
       BlueMaskSize       : Byte;
       BlueFieldPosition  : Byte;
       RsvdMaskSize       : Byte;
       RsvdFieldPosition  : Byte;
       DirectColourMode   : Byte;
       {VESA 2.0 stuff}
       PhysBasePtr        : longint;
       OffScreenMemOffset : pointer;
       OffScreenMemSize   : word;
       paddington: array [49..512] of Byte;
     end;

type DataState=(NONE,OWNER,ALIASING);

type PointType = record
       x,y                : integer
     end;

type PointType3d = record
       x,y,z              : integer
     end;

type TLineList=array[0..MaxScanLines-1] of record xleft,xright:integer end;

type TGfxType = (NOGRAPHIC,LBM,PCX,MSH,PNG,GIF,BMP,TGA,DNY,PAL,COMPTGA,COMPBMP,ILBM);

const TGfxTypeStr:array[TGfxType] of string[3]=('','LBM','PCX','MSH',
                                                   'PNG','GIF','BMP',
                                                   'TGA','DNY','PAL',
                                                   'TGA','BMP','LBM');

type PGfx0 = ^TGfx0;
     TGfx0 = object
       LineOffsets        : array[0..MaxScanLines-1] of longint;
       xRes,yRes          : word;
       GetMaxX,GetMaxY    : word;
       CurDataState       : DataState;
       BPP                : byte;
       _data              : word;    {for 48-bit systems, this is the selector}
       _InternalOffset    : longint; {for 32-bit systems, this will be the ptr}

       constructor init;

       procedure Arc(x_center,y_center,radius,s_angle,e_angle:Word;colour:longint); virtual;
       procedure bar(x1,y1,x2,y2:integer;colour:longint); virtual;
       function BlitSize(const x1,y1,x2,y2:word):longint; virtual;
       procedure BSpline (NumPoints : Word; Var Points : Array Of PointType; Segments : Word; colour:longint); virtual;
       procedure Circle(x_center,y_center,radius:integer;colour:longint); virtual;
       procedure clear(colour:longint); virtual;
       procedure copybuf(DestX,DestY:integer;var SrcObject:Tgfx0;x1,y1,x2,y2:word); virtual;
       procedure CubicBezierCurve (x1,y1,x2,y2,x3,y3,x4,y4:Integer; Segments : Word; colour:longint); virtual;
       procedure Curve (x1,y1,x2,y2,x3,y3:Integer; Segments:Word; colour:longint); virtual;
       procedure DrawPoly(NumPoints:word; var PolyPoints:array of PointType;colour:longint); virtual;
       function ImageSize(x1, y1, x2, y2: word): longint; virtual;
       procedure FillCircle(x_center,y_center,radius:integer;colour:longint); virtual;
       procedure FillConvexPoly(NumPoints:word; var PolyPoints:array of PointType;colour:longint); virtual;
       procedure flip(var SrcObject:TGfx0); virtual;
       procedure GetBlit(const x1,y1,x2,y2:word;blit:pointer48); virtual;
       procedure getimage(x1,y1,x2,y2:word;bitmap:pointer48); virtual;
       function getpixel(x,y:integer):longint; virtual;
       procedure horizline(x1,x2,y:integer;colour:longint); virtual;
       procedure line(x1,y1,x2,y2:integer;colour:longint); virtual;
       procedure putblit(x,y:Integer;blit:pointer48); virtual;
       procedure putimage(x,y:integer;bitmap:pointer48); virtual;
       procedure putpixel(x,y:integer;colour:longint); virtual;
       procedure rawline(x1,y1,x2,y2:word;colour:longint); virtual;
       procedure rectangle(x1,y1,x2,y2:integer;colour:longint); virtual;
       procedure scalebuf(dX1,dY1,dX2,dY2:integer;var SrcObject:TGfx0;sx1,sy1,sx2,sy2:word); virtual;
       procedure vertline(x,y1,y2:integer;colour:longint); virtual;

       destructor done; virtual;
     end;

var RMRegs:record
            case boolean of
            true:(edi,esi,ebp,reserved,ebx,edx,ecx,eax:longint;
                  flags,es,ds,fs,gs,ip,cs,sp,ss:word);
            true:(di,udi,si,usi,bp,ubp,rese,rved,bx,ubx,dx,udx,cx,ucx,ax,uax:word);
           end;

function bswapl(input:longint):longint;
function bswapw(input:word):word;
procedure RealModeInt(IntNum:byte;var CallStruc);
procedure CallRMP(var CallStruc);
function GetSegBaseAddr(Selector:word):longint;
function SegToSelector(SegAddr:word):word;
function ConvertPtr(RMPointer:Pointer):pointer;

implementation

uses newfrontier;

function bswapl(input:longint):longint; assembler;
asm
        mov     ax,word ptr input+2
        xchg    ah,al
        mov     dx,word ptr input
        xchg    dh,dl
end;

function bswapw(input:word):word; assembler;
asm
        mov     ax,input
        xchg    ah,al
end;

{Simulate Real Mode Interrupt}
procedure RealModeInt(IntNum:byte;var CallStruc); assembler;
asm
        mov     ax,300h
        mov     bh,1
        mov     bl,IntNum
        xor     cx,cx
        les     di,CallStruc
        int     31h
end;

{Call Real Mode Procedure (with Far Return Frame)}
procedure CallRMP(var CallStruc); assembler;
asm
        mov     ax,301h
        mov     bh,1
        xor     cx,cx
        les     di,CallStruc
        int     31h
end;

function GetSegBaseAddr(Selector:word):longint; assembler;
asm
        mov     ax,0006h
        mov     bx,Selector
        int     31h
        mov     ax,dx
        mov     dx,cx
end;

function SegToSelector(SegAddr:word):word; assembler;
asm
        mov     ax,0002h
        mov     bx,SegAddr
        int     31h
end;

function ConvertPtr(RMPointer:Pointer):pointer; assembler;
asm
        mov     ax,0002h
        mov     bx,word ptr RMPointer+2
        int     31h
        mov     dx,ax
        mov     ax,word ptr RMPointer
end;

var HLineList:^TlineList;

{clips a line to (0,0),(MaxX,MaxY); returns true if the line segment is in
 bounds.
 Uses HJI's line clipping algorithm}

function ClipLine(var x1,y1,x2,y2:integer;getmaxX,getmaxY:word):boolean;
var tx1,ty1,tx2,ty2:integer;
var outcode:integer;
    andresult,orresult:word;
begin
 ClipLine:=FALSE;
 andresult:=15;
 orresult:=0;
 outcode:=0;
 if x1<0 then inc(outcode,8);
 if x1>getmaxx then inc(outcode,4);
 if y1<0 then inc(outcode,2);
 if y1>getmaxy then inc(outcode);
 andresult:=andresult and outcode;
 orresult:=orresult or outcode;
 outcode:=0;
 if x2<0 then inc(outcode,8);
 if x2>getmaxx then inc(outcode,4);
 if y2<0 then inc(outcode,2);
 if y2>getmaxy then inc(outcode);
 andresult:=andresult and outcode;
 orresult:=orresult or outcode;
 if andresult>0 then exit; {line is completely off the screen}
 if orresult=0 then
  begin
   ClipLine:=TRUE;
   exit
  end;
 {preset the values in case they're not changed}
 tx1:=x1;
 ty1:=y1;
 tx2:=x2;
 ty2:=y2;
 if (x1<0) then
  begin
   ty1:=(x2*longint(y1)-x1*longint(y2)) div (x2-x1);
   tx1:=0;
  end else
 if (x2<0) then
  begin
   ty2:=(x2*longint(y1)-x1*longint(y2)) div (x2-x1);
   tx2:=0;
  end;

 if (x1>GetMaxX) then
  begin
   ty1:=(longint(y1)*(x2-getmaxx)+longint(y2)*(getmaxx-x1)) div (x2-x1);
   tx1:=getmaxx;
  end
 else
 if (x2>GetMaxX) then
  begin
   ty2:=(longint(y1)*(x2-getmaxx)+longint(y2)*(GetMaxX-x1)) div (x2-x1);
   tx2:=GetmaxX;
  end;

 if ((ty1<0) and (ty2<0)) or ((ty1>getmaxy) and (ty2>getmaxy)) then exit;

 if (ty1<0) then
  begin
   tx1:=(x1*longint(y2)-x2*longint(y1)) div (y2-y1);
   ty1:=0;
  end else
 if (ty2<0) then
  begin
   tx2:=(x1*longint(y2)-x2*longint(y1)) div (y2-y1);
   ty2:=0;
  end;

 if (ty1>GetmaxY) then
  begin
   tx1:=(longint(x1)*(y2-getmaxy)+longint(x2)*(GetmaxY-y1)) div (y2-y1);
   ty1:=GetMaxY
  end
 else
 if (ty2>GetMaxY) then
  begin
   tx2:=(longint(x1)*(y2-getmaxy)+longint(x2)*(GetmaxY-y1)) div (y2-y1);
   ty2:=GetMaxY
  end;
 if (tx1<0) or (tx2<0) or (tx1>getmaxx) or (tx2>getmaxx) then exit;
 ClipLine:=TRUE;
 x1:=tx1;
 y1:=ty1;
 x2:=tx2;
 y2:=ty2;
end;

{clips a line to (0,0),(MaxX,MaxY); returns true if the line segment is on
 the screen.}

function ClipLineHoriz(var x1,y1,x2,y2:integer;getmaxY:word):boolean;
var tx1,ty1,tx2,ty2:integer;
var outcode:integer;
    andresult,orresult:word;
begin
 ClipLineHoriz:=FALSE;
 andresult:=15;
 orresult:=0;
 outcode:=0;
 if y1<0 then inc(outcode,2);
 if y1>getmaxy then inc(outcode);
 andresult:=andresult and outcode;
 orresult:=orresult or outcode;
 outcode:=0;
 if y2<0 then inc(outcode,2);
 if y2>getmaxy then inc(outcode);
 andresult:=andresult and outcode;
 orresult:=orresult or outcode;
 if andresult>0 then exit; {line is completely off the screen}
 if orresult=0 then
  begin
   ClipLineHoriz:=TRUE;
   exit
  end;
 {preset the values in case they're not changed}
 tx1:=x1;
 ty1:=y1;
 tx2:=x2;
 ty2:=y2;
 if ((ty1<0) and (ty2<0)) or ((ty1>getmaxy) and (ty2>getmaxy)) then exit;

 if (ty1<0) then
  begin
   tx1:=(x1*longint(y2)-x2*longint(y1)) div (y2-y1);
   ty1:=0;
  end else
 if (ty2<0) then
  begin
   tx2:=(x1*longint(y2)-x2*longint(y1)) div (y2-y1);
   ty2:=0;
  end;

 if (ty1>GetmaxY) then
  begin
   tx1:=(longint(x1)*(y2-getmaxy)+longint(x2)*(GetmaxY-y1)) div (y2-y1);
   ty1:=GetMaxY
  end
 else
 if (ty2>GetMaxY) then
  begin
   tx2:=(longint(x1)*(y2-getmaxy)+longint(x2)*(GetmaxY-y1)) div (y2-y1);
   ty2:=GetMaxY
  end;
 ClipLineHoriz:=TRUE;
 x1:=tx1;
 y1:=ty1;
 x2:=tx2;
 y2:=ty2;
end;


{This is how the fillpoly works:
Fillpoly preinitializes an array of xleft and xright values to
xleft=-1, xright=getmaxx+1
Using line clipping, it will compute the xleft and xright values for
each line segment in the polygon.
The line plotter will use the x value it wants to plot in this way:

If xval<xleft then xleft:=xval else if xval>xright then xright:=xval;

Since -1 will be interpreted as $ffff, any value on the screen will be less
and therefore the -1 (or $ffff in this case) will be replaced.
Horizline will clip -1 to 0, and getmaxx+1 to getmaxx, thus getting scan
lines that didn't have any part of the polygon on the screen.

}
var Reverse:boolean;
procedure ScanEdge(a1,b1,a2,b2:integer);
{uses the Symmetric Double Step Line Algorithm from graphics gems 1}
var dx,dy,incr1,incr2,D,x,y,xend,c,pixels_left:integer;
var x1,y1:integer;
var sign_x,sign_y,step,i:integer;

procedure plot(x,y:integer); assembler;
asm
        les     di,hlinelist
        mov     bx,y
        mov     ax,x
        cmp     reverse,FALSE
        je      @@norev
        xchg    ax,bx
@@norev:
        shl     bx,2
        add     di,bx
        cmp     ax,es:[di]
        jge     @@noxleft
        mov     es:[di],ax
@@noxleft:
        cmp     ax,es:[di+2]
        jle     @@noxright
        mov     es:[di+2],ax
@@noxright:
end; {
begin
 if reverse then
  with HLineList^[x] do
   begin
    if y<0 then begin if y<xleft then xleft:=y end else
     if y<word(xleft) then xleft:=y;
    if y>xright then xright:=y
   end
 else
  with HLineList^[y] do
  begin
   if x<0 then begin if x<xleft then xleft:=x end else
    if x<word(xleft) then xleft:=x;
   if x>xright then xright:=x
  end
end;  }

begin
 Reverse:=FALSE; {by default}

 if (a1=a2) then
  begin
   x:=a1;
   if b1>b2 then
    begin
     y:=b1;
     b1:=b2;
     b2:=y;
    end;
   for y:=b1 to b2 do
    plot(x,y);
   exit;
  end;

 dx:=(a2-a1);
 if (dx<0) then
  begin
   dx:=-dx;
   sign_x:=-1
  end
 else sign_x:=1;

 dy:=(b2-b1);
 if (dy<0) then
  begin
   dy:=-dy;
   sign_y:=-1
  end
 else sign_y:=1;

 if (sign_x=sign_y) then step:=1 else step:=-1;

 {note: reverse is already set to false}
 if (dy>dx) then
  asm {doing this in asm is so much faster than using a swap function}
        mov     ax,a1
        xchg    ax,b1
        mov     a1,ax

        mov     ax,a2
        xchg    ax,b2
        mov     a2,ax

        mov     ax,&dx
        xchg    ax,dy
        mov     &dx,ax
        mov     reverse,TRUE
  end;

 if (a1>a2) then
  begin
   x:=a2;
   y:=b2;
   x1:=a1;
   y1:=b1
  end
 else
  begin
   x:=a1;
   y:=b1;
   x1:=a2;
   y1:=b2
  end;

 xend:=(dx-1) shr 2;
 pixels_left:=(dx-1) and 3;

 plot(x,y);
 plot(x1,y1);
 incr2:=(4*dy) - (2*dx);
 if (incr2<0) then
  begin
   c:=2*dy;
   incr1:=2*c;
   D:=incr1-dx;

   for i:=0 to xend-1 do
    begin
     inc(x);
     dec(x1);
     if (d<0) then
      begin
       plot(x,y);
       inc(x);
       plot(x,y);
       plot(x1,y1);
       dec(x1);
       plot(x1,y1);
       inc(d,incr1);
      end
     else
      begin  {pattern 2}
       if (d<c) then
        begin
         plot(x,y);
         inc(x);
         inc(y,step);
         plot(x,y);
         plot(x1,y1);
         dec(x1);
         dec(y1,step);
         plot(x1,y1);
        end
       else
        begin {pattern 3 forward}
         inc(y,step);
         plot(x,y);
         inc(x);
         plot(x,y);
         dec(y1,step);
         plot(x1,y1);
         dec(x1);
         plot(x1,y1)
        end;
       inc(d,incr2);
      end;
    end; {for}
   if (pixels_left<>0) then
    begin
     if (D<0) then
      begin
       plot(x+1,y);
       if (pixels_left>1) then plot(x+2,y);
       if (pixels_left>2) then plot(x1-1,y1);
      end
     else
      begin
       if (d<c) then
        begin {pattern 2}
         plot(x+1,y);
         if (pixels_left>1) then plot(x+2,y+step);
         if (pixels_left>2) then plot(x1-1,y1);
        end
       else
        begin {pattern 3}
         plot(x+1,y+step);
         if (pixels_left>1) then plot(x+2,y+step);
         if (pixels_left>2) then plot(x1-1,y1-step)
        end;
      end;
    end {if pixels_left<>0}
  end {if incr2<0)}
 else {slope greater than 1/2}
  begin
   c:=2*(dy-dx);
   incr1:=2*c;
   d:=incr1+dx;
   for i:=0 to xend-1 do
    begin
     inc(x);
     dec(x1);
     if (d>0) then
      begin {pattern 4}
       inc(y,step);
       plot(x,y);
       inc(x);
       inc(y,step);
       plot(x,y);
       dec(y1,step);
       plot(x1,y1);
       dec(x1);
       dec(y1,step);
       plot(x1,y1);
       inc(d,incr1);
      end {if (d>0)}
     else
      begin
       if (D<c) then
        begin {pattern 2 forwards}
         plot(x,y);
         inc(x);
         inc(y,step);
         plot(x,y);
         {pattern 2 backwards}
         plot(x1,y1);
         dec(x1);
         dec(y1,step);
         plot(x1,y1)
        end {if (d<c)}
       else
        begin {pattern 3 forwards}
         inc(y,step);
         plot(x,y);
         inc(x);
         plot(x,y);
         {pattern 3 backwards}
         dec(y1,step);
         plot(x1,y1);
         dec(x1);
         plot(x1,y1)
        end;
       inc(d,incr2);
      end; {else}
    end; {for}
   if (pixels_left<>0) then
    begin
     if (D>0) then
      begin      {pattern 4}
       inc(y,step);
       plot(x+1,y);
       if (pixels_left>1) then plot(x+2,y+step);
       if (pixels_left>2) then plot(x1-1,y1-step);
      end {if (d>0)}
     else
      begin
       if (D<c) then
        begin  {pattern 2}
         plot(x+1,y);
         if (pixels_left>1) then plot(x+2,y+step);
         if (pixels_left>2) then plot(x1-1,y1);
        end {if (D<c)}
       else
        begin {pattern 3}
         plot(x+1,y+step);
         if (pixels_left>1) then plot(x+2,y+step);
         if (pixels_left>2) then
          if (d>c) then
           plot(x1-1,y1-step)
          else plot(x1-1,y1);
        end;
      end {else}
    end {if pixels_left<>0}
  end
end;

constructor TGfx0.init;
begin
 _Data:=0;
 _InternalOffset:=0;
 BPP:=0;
 xRes:=0;
 yRes:=0;
 GetMaxX:=0;
 GetMaxY:=0;
 fillchar(LineOffsets,sizeof(lineoffsets),0);
 CurDataState:=NONE
end;

Procedure TGfx0.Arc;
{ An algorithm to draw an arc. Crude but it works (anyone have a better one?) }
Var
  p : Integer;
  x, y : Word;
  Alpha : Real;

Begin
  If radius=0 then
   Begin
    PutPixel(x_center,y_center,colour);
    Exit
   End;
  s_angle:=s_angle MOD 361;
  e_angle:=e_angle MOD 361;
  If s_angle>e_angle then
   Begin
    s_angle:=s_angle Xor e_angle;
    e_angle:=e_angle Xor s_angle;
    s_angle:=e_angle Xor s_angle
   End;
  x:=0;
  y:=Radius;
  p:=3-2*Radius;
  While x<=y Do
    Begin
      Alpha:=RadToDeg*Arctan (x/y);
      If (Alpha>=s_angle) And (Alpha<=e_angle) then PutPixel (x_center-x, y_center-y, Colour);
      If (90-Alpha>=s_angle) And (90-Alpha<=e_angle) then PutPixel(x_center-y, y_center-x,colour);
      If (90+Alpha>=s_angle) And (90+Alpha<=e_angle) then PutPixel(x_center-y, y_center+x,colour);
      If (180-Alpha>=s_angle) And (180-Alpha<=e_angle) then PutPixel(x_center-x, y_center+y,colour);
      If (180+Alpha>=s_angle) And (180+Alpha<=e_angle) then PutPixel(x_center+x, y_center+y,Colour);
      If (270-Alpha>=s_angle) And (270-Alpha<=e_angle) then PutPixel(x_center+y, y_center+x,Colour);
      If (270+Alpha>=s_angle) And (270+Alpha<=e_angle) then PutPixel(x_center+y, y_center-x,Colour);
      If (360-Alpha>=s_angle) And (360-Alpha<=e_angle) then PutPixel(x_center+x, y_center-y,Colour);
      If p<0 then inc(p,4*x+6) Else
        Begin
          p:=p+4*(x-y)+10;
          Dec (y)
        End;
      Inc (x)
    End
End;

procedure TGfx0.bar;
var yy:word;
    temp:integer;
begin
 if x2<x1 then
  begin
   temp:=x2;
   x2:=x1;
   x1:=temp
  end;
 if y2<y1 then
  begin
   temp:=y2;
   y2:=y1;
   y1:=temp
  end;
 {Horizline automagically clips on the left and right :}
 if y1<0 then y1:=0;
 if y2>getmaxy then y2:=getmaxy;
 for yy:=y1 to y2 do
  horizline(x1,x2,yy,colour)
end;

function TGfx0.BlitSize;
var x,y,tmploc:word; {Blits are limited to 64K?}
    zerocount:byte;
    InZeros:boolean;
    xlcount:word;
    height:word;
begin
 height:=(y2-y1)+1;
 tmploc:=0;
 for y:=y1 to y2 do
  begin
   zerocount:=0;
   inc(tmploc);            {Save room for xlcount}
   x:=x1;
   inzeros:=(getpixel(x,y)=TransparentColour);  {Should we start in zeros?}
   while x<=x2 do           {Continue until we are done with the line}
    case inzeros of
     true:begin
           zerocount:=0;   {How many zeros?}
           while (x<=x2) and (getpixel(x,y)=TransparentColour) do
            begin
             inc(x);
             inc(zerocount)
            end;
           inzeros:=FALSE  {we are no longer in zeros}
          end;
     false:begin
            inc(tmploc,4); {zerocount (offset):word; xlcount:word}
            repeat
             inc(x);
             {NOTE!!! The next line presumes at least 8bpp.  If there
              is an implementation of less than 8bpp you will need to
              re-create this function!!!!}

              {It's BPP+7 so that 4bpp returns 1, and 15bpp and 16bpp return 2}
             inc(tmploc,(BPP+7) shr 3);
            until (x>=x2) or (getpixel(x,y)=TransparentColour);
            inzeros:=True
           end
   end
  end;
 Blitsize:=tmploc+Height*2+8;
end;

Procedure TGfx0.BSpline;
{ Draw a BSpline approximating a curve defined by the array of points. }
{ Beware! A B-Spline generaly does not normally pass through the points}
{ defining it !                                                        }
 Function Calculate (mu : Real; p0, p1, p2, p3 : Integer) : Integer;
 Var mu2, mu3 : Real;

 Begin
  mu2:=mu*mu;
  mu3:=mu2*mu;
  Calculate:=Round ((1/6)*(mu3*(-p0+3*p1-3*p2+p3)+
                           mu2*(3*p0-6*p1+3*p2)+
                           mu *(-3*p0+3*p2)+(p0+4*p1+p2)));
 End;

Var
  mu, mudelta : Real;
  x1, y1, x2, y2, n, h : Integer;

Begin
  If (NumPoints<4) Or (NumPoints>16383) then Exit;
  mudelta:=1/Segments;
  For n:=3 to NumPoints-1 Do
   Begin
    mu:=0;
    x1:=Calculate (mu,Points[n-3].x,Points[n-2].x,Points[n-1].x,Points[n].x);
    y1:=Calculate (mu,Points[n-3].y,Points[n-2].y,Points[n-1].y,Points[n].y);
    mu:=mu+mudelta;
    For h:=1 to Segments Do
     Begin
      x2:=Calculate (mu,Points[n-3].x,Points[n-2].x,Points[n-1].x,Points[n].x);
      y2:=Calculate (mu,Points[n-3].y,Points[n-2].y,Points[n-1].y,Points[n].y);
      Line(x1,y1,x2,y2,colour);
      mu:=mu+mudelta;
      x1:=x2;
      y1:=y2
    End
  End
End;

Procedure TGfx0.Circle;
{ Draw a circle with center (x_center,y_center) and radius 'radius' }
Var
  x, y, d : Integer;
Begin
  { bressenham circle algorithm using integer-only arithmetic }
  x:=0;
  y:=radius;
  d:=2*(1-radius);
  While y>=0 Do
   Begin
    PutPixel (x_center+x,y_center+y,Colour);
    PutPixel (x_center+x,y_center-y,Colour);
    PutPixel (x_center-x,y_center+y,Colour);
    PutPixel (x_center-x,y_center-y,Colour);
    If d + y > 0 then
     Begin
      Dec (y);
      Dec (d,2*y+1)
     End;
    If x > d then
     Begin
      Inc (x);
      Inc (d,2*x+1)
     End
   End
End;

procedure TGfx0.Clear;
begin
end;

procedure TGfx0.CopyBuf;
begin
end;

Procedure TGfx0.CubicBezierCurve;
{ Draw a cubic bezier-curve directly using the basis functions }
Var
  tx1, tx2, tx3, ty1, ty2, ty3, mu, mu2, mu3, mudelta : Real;
  xstart, ystart, xend, yend, n : Integer;

Begin
  If (Segments<1) then Exit;
  If Segments>128 then Segments:=128;

  mudelta:=1/Segments;
  mu:=0;
  tx1:=-x1+3*x2-3*x3+x4; ty1:=-y1+3*y2-3*y3+y4;
  tx2:=3*x1-6*x2+3*x3;   ty2:=3*y1-6*y2+3*y3;
  tx3:=-3*x1+3*x2;       ty3:=-3*y1+3*y2;

  xstart:=x1;
  ystart:=y1;
  mu:=mu+mudelta;
  For n:=1 to Segments Do
   Begin
    mu2:=mu*mu;
    mu3:=mu2*mu;
    xend:=Round (mu3*tx1+mu2*tx2+mu*tx3+x1);
    yend:=Round (mu3*ty1+mu2*ty2+mu*ty3+y1);
    Line (xstart, ystart, xend, yend,colour);
    mu:=mu+mudelta;
    xstart:=xend;
    ystart:=yend
  End
End;

Procedure TGfx0.Curve;
{ Draw a curve from (x1,y1) through (x2,y2) to (x3,y3) divided in 'Segments' segments }
Var
  lsteps, ex, ey, fx, fy : LongInt;
  t1, t2 : Integer;

Begin
  x2:=(x2 SHL 1)-((x1+x3) SHR 1);
  y2:=(y2 SHL 1)-((y1+y3) SHR 1);
  lsteps:=Segments;
  If (lsteps<2) then lsteps:=2;
  If (lsteps>128) then lsteps:=128;  { Clamp value to avoid overcalculation }
  ex:=(LongInt (x2-x1) SHL 17) DIV lsteps;
  ey:=(LongInt (y2-y1) SHL 17) DIV lsteps;
  fx:=(LongInt (x3-(2*x2)+x1) SHL 16) DIV (lsteps*lsteps);
  fy:=(LongInt (y3-(2*y2)+y1) SHL 16) DIV (lsteps*lsteps);
  Dec (lsteps);
  While lsteps>0 Do Begin
    t1:=x3;
    t2:=y3;
    x3:=(((fx*lsteps+ex)*lsteps) SHR 16)+x1;
    y3:=(((fy*lsteps+ey)*lsteps) SHR 16)+y1;
    Line (t1,t2,x3,y3,colour);
    Dec (lsteps);
  End;
  Line (x3,y3,x1,y1,colour);
End;

procedure TGfx0.DrawPoly;
var count:word;
begin
 case NumPoints of
  1:putpixel(PolyPoints[0].x,PolyPoints[0].y,colour);
  2:line(PolyPoints[0].x,PolyPoints[0].y,PolyPoints[1].x,PolyPoints[1].y,colour)
 else
  for count:=0 to NumPoints-1 do
   line(PolyPoints[count].x,PolyPoints[count].y,
        PolyPoints[(count+1) mod NumPoints].x,
        PolyPoints[(count+1) mod NumPoints].y,colour)
 end
end;

Procedure TGfx0.FillCircle;
{ Draw a circle with center (x_center,y_center) and radius 'radius' }
Var
  x, y, d : Integer;
  ystart,yend:integer;
  count:word;

Begin
  { bressenham circle algorithm using integer-only arithmetic }
  x:=0;
  y:=radius;
  d:=2*(1-radius);
  ystart:=y_center-radius;
  yend:=y_center+radius;

  if ystart<0 then ystart:=0;
  if yend>getmaxy then yend:=getmaxy;
{  for count:=ytop to ybot do
   with HLineList^[count] do
    begin
     xleft:=x_center;
     xright:=x_center;
    end;
}
  While y>=0 Do
   Begin
    if (y_center+y<=GetMaxY) then
    with HLineList^[y_center+y] do
     begin
      xleft:=x_center-x;
      xright:=x_center+x;
     end;

    if (y_center-y>=0) then
    with HLineList^[y_center-y] do
     begin
      xleft:=x_center-x;
      xright:=x_center+x;
     end;
    If d + y > 0 then
     Begin
      Dec (y);
      Dec (d,2*y+1)
     End;
    If x > d then
     Begin
      Inc (x);
      Inc (d,2*x+1)
     End
   End;

 for count:=ystart to yend do
  with hlinelist^[count] do
   horizline(xleft,xright,count,colour)

End;

procedure TGfx0.FillConvexPoly;
var outcode:integer;
var count:integer;
    andresult,orresult:word;
    firstinb:integer;
    ystart,yend,xstart,xend:integer;
    temp:word;
    x1,y1,x2,y2:integer;
begin
 if (NumPoints=0) then exit; (* reject null polygons *)
 andresult:=15;
 firstinb:=-1;
 orresult:=0;
 XStart:=PolyPoints[0].x;
 ystart:=PolyPoints[0].y;
 xend:=xstart;
 yend:=yStart;
 for count:=0 to NumPoints-1 do
  with PolyPoints[count] do
  begin
   outcode:=0;
   if x<0 then inc(outcode,8);
   if x>getmaxx then inc(outcode,4);
   if y<0 then inc(outcode,2);
   if y>getmaxy then inc(outcode);
   andresult:=andresult and outcode;
   orresult:=orresult or outcode;
   if Y<yStart then yStart:=Y else if Y>yend then yend:=Y;
   if X<xStart then xStart:=x else if X>xend then xend:=x;
  end;
 if andresult>0 then exit; {polygon is completely off the screen}
 if yend>getmaxy then yend:=getmaxy;
 if ystart<0 then ystart:=0;

 asm
        mov     bx,ystart
        les     di,hlinelist
        shl     bx,2
        add     di,bx
        mov     ax,xStart
        push    ax
        mov     cx,yend
        db 66h; shl ax,16
        sub     cx,ystart
        pop     ax
        inc     cx
        mov     ax,xend
        db 66h; rep stosw
 end;
 for count:=0 to NumPoints-1 do
  begin
   x1:=PolyPoints[count].X;
   y1:=PolyPoints[count].Y;
   x2:=PolyPoints[(count+1) mod NumPoints].X;
   y2:=PolyPoints[(count+1) mod NumPoints].Y;
   if ClipLineHoriz(x1,y1,x2,y2,getmaxy) then ScanEdge(x1,y1,x2,y2);
  end;

 for count:=ystart to yend do
  with hlinelist^[count] do
   horizline(xleft,xright,count,colour)
end;

procedure TGfx0.Flip;
var count,xCount,yCount:word;
begin
 if SrcObject._data=0 then exit;
 xCount:=SrcObject.GetMaxX+1;
 if xCount>GetMaxX+1 then xCount:=GetMaxX+1;
 yCount:=SrcObject.GetMaxY;
 if yCount>GetMaxY then yCount:=GetMaxY;
(* if SrcObject._data=Sega000 then {This is a vesa8 object, so we need}
  begin                          {to typecast the gfx8 object to a vesa8 obj}
  {(@SrcObject)^ is a dereferenced pointer to the object.  It didn't
   work any other way }
   for count:=0 to yCount do
    Tscreen0((@SrcObject)^).CopyLineFromScreen(0,count,
               ptr48(_data,_internalOffset+LineOffsets[count]),xCount)
  end
 else  *)
 for count:=0 to yCount do
  move32(ptr48(SrcObject._data,SrcObject._InternalOffset+SrcObject.LineOffsets[count]),
         ptr48(_data,_internalOffset+LineOffsets[count]),
         xCount)
end;

procedure TGfx0.getblit;
begin
end;

procedure TGfx0.getimage;
begin
end;

function TGfx0.getpixel;
begin
end;

procedure TGfx0.horizline;
begin
end;

function TGfx0.ImageSize;
begin
 ImageSize:=((abs(x2-x1)+1)*longint(abs(y2-y1)+1)*((BPP+7) shr 3)+4);
end;

procedure TGfx0.line;
begin
 {clipline modifies (x1,y1),(x2,y2) and returns true or returns false if the
  line is outside the buffer}
 if ClipLine(x1,y1,x2,y2,getmaxx,getmaxy) then rawline(x1,y1,x2,y2,colour)
end;

procedure TGfx0.putblit;
begin
end;

procedure TGfx0.putimage;
begin
end;

procedure TGfx0.putpixel;
begin
end;

procedure TGfx0.rawline;
begin
end;

procedure TGfx0.rectangle;
begin
 horizline(x1,x2,y1,colour); {Horizline has built in clipping}
 vertline(x2,y1,y2,colour);  {vertline has built in clipping too}
 vertline(x1,y1,y2,colour);
 horizline(x1,x2,y2,colour)
end;

procedure TGfx0.scalebuf;
begin
end;

procedure TGfx0.vertline;
begin
end;

destructor TGfx0.done;
begin
 if (CurDataState=OWNER) then freemem32(_data);
end;

procedure GetVidMode; assembler;
asm
        mov     ax,4f03h
	int     $10
	mov     OldVidMode,bx
end;

var SaveExit: Pointer;

procedure ObjGfxExit; far;
begin
  ExitProc := SaveExit;
  Dispose(hLineList);
end; { ObjGfxExit }

begin
 SaveExit := ExitProc;
 ExitProc := @ObjGfxExit;
 new(hLineList); {used for fillpolys; takes up about 5K of heap space}
 GetVidMode;
 SegC000:=SegToSelector($c000);
 Seg0000:=SegToSelector($0000);
 Seg0040:=SegToSelector($0040);
 SegA000:=SegToSelector($A000);
 SegF000:=SegToSelector($F000);
end.