(********************************************************************)
(*                                                                  *)
(*                               OBJGFX8.PAS                        *)
(*                                                                  *)
(*                     Object Oriented Graphics Library             *)
(*                                                                  *)
(*                                                                  *)
(*                            Copyright (c) 1999                    *)
(*                      Mark J Iuzzolino, Daniel Robbins            *)
(*                            All Rights Reserved                   *)
(*                                                                  *)
(*                                                                  *)
(*                                                                  *)
(********************************************************************)

(*
 *   Rev 2.0  1999/05/24 mji
 * Converted ObjGfx16 to work with the new changes.
 *
 *   Rev 1.0  1998/09/21 mji
 * Initial Revision
 *)

(*
 * Presumed data types and sizes (for all you non-pascal programmers):
 *
 * boolean   : 8-bit boolean value
 * wordbool  : 16-bit boolean value
 * byte      : unsigned 8-bit value
 * shortint  : signed 8-bit value
 * char      : unsigned 8-bit value (character)
 * word      : unsigned 16-bit value
 * integer   : signed 16-bit value
 * longint   : signed 32-bit value
 * real      : 48-bit floating point
 * pointer   : 32-bit pointer (16-bit selector; 16-bit offset)
 * pointer48 : 48-bit pointer (16-bit selector; 32-bit offset)
 * string    : array of char; string length is string[0]
 *
 *)

{$A+,B-,D+,E+,F-,G+,I+,L+,N-,P-,Q-,R-,S+,T-,V+,X+,Y+}

unit ObjGfx16;

{$IFNDEF DPMI}
 -- Protected mode *only*
{$ENDIF}

interface

uses ObjGfx0,ObjGfx8;

const _320x200x15bpp       = $10D;
const _640x480x15bpp       = $110;
const _800x600x15bpp       = $113;
const _1024x768x15bpp      = $116;
const _1280x1024x15bpp     = $119;
const _320x200x16bpp       = $10E;
const _640x480x16bpp       = $111;
const _800x600x16bpp       = $114;
const _1024x768x16bpp      = $117;
const _1280x1024x16bpp     = $11A;

type PGfx16=^Tgfx16;
     Tgfx16=object(TGfx0)
       procedure AdoptGfx8(var Gfx8Object:TGfx8);
       procedure AliasGfx16(var Gfx16Object:TGfx16;x1,y1,x2,y2:word); virtual;
       procedure Clear(colour:longint); virtual;
       procedure CopyBuf(DestX,DestY:integer;var SrcObject:Tgfx0;x1,y1,x2,y2:word); virtual;
       procedure Flip(var SrcObject:TGfx0); virtual;
       procedure FlipX; virtual;
       procedure FlipY; virtual;
       procedure horizline(x1,x2,y:integer;colour:longint); virtual;
       procedure getblit(const x1,y1,x2,y2:word;blit:pointer48); virtual;
       function GetMem(myw,myh:word):boolean;
       procedure GetImage(x1,y1,x2,y2:word;bitmap:pointer48); virtual;
       function GetPixel(x,y:integer):longint; virtual;
{       function LoadGfx(const gfxfile:string):boolean;}
       procedure putblit(x,y:Integer;blit:pointer48); virtual;
       procedure putimage(x,y:integer;bitmap:pointer48); virtual;
       procedure PutPixel(x,y:integer;colour:longint); virtual;
       procedure RawLine(x1,y1,x2,y2:word;colour:longint); virtual;
{       function SaveGfx(const gfxfile:string;GfxType:TGfxType):boolean;}
       procedure scalebuf(dX1,dY1,dX2,dY2:integer;var SrcObject:Tgfx0;sx1,sy1,sx2,sy2:word); virtual;
       procedure vertline(x,y1,y2:integer;colour:longint); virtual;
     end;

type  PScreen16=^TScreen16;
      TScreen16=object(Tgfx16)
       VESARec            : ^TVESA_Rec;
       ModeRec            : ^TMode_Rec;
       SwitchBank         : procedure;
       _BankVals          : array[0..479] of word;
       _CurWriteBank      : word;
       _CurReadBank       : word;
       _ReadWindow        : word;
       _VESASelector      : word;
       _CodeSelector      : word;
       _InGraphics        : boolean;
       _LFB               : boolean;

       constructor init;
       procedure Clear(colour:longint); virtual;
       procedure copybuf(DestX,DestY:integer;var SrcObject:Tgfx0;x1,y1,x2,y2:word); virtual;
       procedure CopyLineFromScreen(x,y:word;src:pointer48;size:word);
       procedure CopyLineToScreen(src:pointer48;x,y:word;size:word);
       procedure GetImage(x1,y1,x2,y2:word;bitmap:pointer48); virtual;
       function GetModeInfo(mode:word):word;
       function GetPixel(x,y:integer):longint; virtual;
       procedure horizline(x1,x2,y:integer;colour:longint); virtual;
       function IsModeAvailable(ModeNum:word):boolean;
       function IsVesaInstalled:word;
       procedure flip(var SrcObject:TGfx0); virtual;
       procedure putblit(x,y:Integer;blit:pointer48); virtual;
       procedure putimage(x,y:integer;bitmap:pointer48); virtual;
       procedure PutPixel(x,y:integer;colour:longint); virtual;
       procedure RawLine(x1,y1,x2,y2:word;colour:longint); virtual;
       procedure scalebuf(dX1,dY1,dX2,dY2:integer;var SrcObject:TGfx0;sx1,sy1,sx2,sy2:word); virtual;
       function SetMode(mode:word):boolean;
       procedure SetupMode(mode:word);
       function UsingLFB:boolean;
       procedure vertline(x,y1,y2:integer;colour:longint); virtual;
       procedure WaitForRetrace;
       destructor done; virtual;
     end;

var    RedShifter         : Byte;
var    GreenShifter       : Byte;
var    BlueShifter        : Byte;
var    RedFieldPosition   : Byte;
var    GreenFieldPosition : Byte;
var    BlueFieldPosition  : Byte;

function RGB(red,green,blue:byte):longint;
function HSL(h,s,l:real):longint;

implementation
uses dos,winapi,newfrontier;

function HSL(h,s,l:real):longint;
var r,g,b:real;
    m1,m2:real;
function HuetoRGB(m1,m2,h:real):real;
begin
 if (h<0) then h:=h+1.0;
 if (h>1) then h:=h-1.0;
 if (6.0*h<1) then HueToRGB:=(m1+(m2-m1)*h*6.0)
 else
  if (2.0*h<1) then HueToRGB:=m2
  else
   if (3.0*h<2.0) then
    HueToRGB:=(m1+(m2-m1)*((2.0/3.0)-h)*6.0)
   else HueToRGB:=m1
end;

begin
 if (S=0) then
  begin
   r:=L;
   g:=L;
   b:=L
  end
 else
  begin
   if (L<=0.5) then
    m2:=l*(1.0+S)
   else
    m2:=L+S-L*S;
   m1:=2.0*L-m2;
   r:=HueToRGB(m1,m2,H+1.0/3.0);
   g:=HueToRGB(m1,m2,H);
   b:=HueToRGB(m1,m2,H-1.0/3.0);
 end;
 HSL:=RGB(trunc(r)*255,trunc(g)*255,trunc(b)*255);
end;

procedure TGfx16.AdoptGfx8;
var PixTab:array[byte] of word;
var xx,yy:word;
    idx:byte;
begin
 if (CurDataState=NONE) then
  if not getmem(Gfx8Object.GetMaxX+1,Gfx8Object.GetMaxY+1) then exit;

 with Gfx8Object do
  for idx:=0 to 255 do
   with Pal[idx] do
    PixTab[idx]:=RGB(r shl 2,g shl 2,b shl 2);
 for yy:=0 to Gfx8Object.GetMaxY do
  for xx:=0 to Gfx8Object.GetMaxX do
   putpixel(xx,yy,PixTab[Gfx8OBject.getpixel(xx,yy)])
end;

procedure TGfx16.AliasGfx16;
var count,width,height:word;
begin
 if CurDataState<>OWNER then
  begin
   if x2<x1 then
    begin
     count:=x2;
     x2:=x1;
     x1:=count
    end;
   if y2<y1 then
    begin
     count:=y2;
     y2:=y1;
     y1:=count
    end;
   width:=(x2-x1)+1;
   height:=(y2-y1)+1;
   if height>MaxScanLines then height:=MaxScanLines;
   CurDataState:=ALIASING;
   Xres:=Gfx16Object.Xres;
   YRes:=Gfx16Object.YRes;
   GetMaxX:=width-1;
   GetMaxY:=height-1;
   _data:=Gfx16Object._Data;
   LineOffsets[0]:=0;
   for Count:=1 to height-1 do
    LineOffsets[Count]:=LineOffsets[Count-1]+Xres;
   _InternalOffset:=Gfx16Object.LineOffsets[y1]+(x1*2)+Gfx16Object._InternalOffset;
  end
end;

procedure TGfx16.FlipX;
var xwidth,count,buf:word;
begin
 xwidth:=(GetMaxX+1)*2;
 getmem32(buf,XWidth);
 for count:=0 to (GetMaxY div 2) do
  begin
   move32(ptr48(_data,_InternalOffset+LineOffsets[count]),ptr48(buf,0),Xwidth);
   move32(ptr48(_data,_InternalOffset+LineOffsets[GetMaxY-count]),
          ptr48(_data,_InternalOffset+LineOffsets[count]),xwidth);
   move32(ptr48(buf,0),ptr48(_data,_InternalOffset+LineOffsets[GetMaxY-count]),xwidth);
  end;
 freemem32(buf)
end;

procedure TGfx16.GetBlit;
var Offset:longint;
    Selector,x,y,idx:word;
    zerocount:word;
    found,InZeros:boolean;
    LineCountIdx,xlcountIdx:word;
    StartX,StartY,EndX,Endy:word; {relative pic coords}
begin
 Found:=FALSE;
 StartX:=x1;
 StartY:=y1;
 EndX:=x2;
 EndY:=y2;
 while not(found) and (StartX<x2) do
  begin
   for y:=y1 to y2 do
    found:=found or (getpixel(StartX,y)<>TransparentColour);
   if not found then inc(StartX)
  end;
 found:=FALSE;
 while not(found) and (StartY<y2) do
  begin
   for x:=StartX to x2 do
    found:=found or (getpixel(x,StartY)<>TransparentColour);
   if not found then inc(StartY);
  end;
 found:=FALSE;
 while not(found) and (EndX>StartX) do
  begin
   for y:=StartY to y2 do
    found:=found or (getpixel(EndX,y)<>TransparentColour);
   if not found then dec(EndX);
  end;
 found:=FALSE;
 while not(found) and (EndY>StartY) do
  begin
   for x:=StartX to EndX do
    found:=found or (getpixel(x,EndY)<>TransparentColour);
   if not found then dec(EndY);
  end;
 Selector:=seg48(blit);
 Offset:=ofs48(blit);
 WriteMem32w(ptr48(Selector,0),StartX-x1);
 WriteMem32w(ptr48(Selector,2),StartY-y1);
 WriteMem32w(ptr48(Selector,4),EndX-x1);
 WriteMem32w(ptr48(Selector,6),EndY-y1);
 inc(offset,8);
 idx:=((y2-y1)+1)*2;
 for y:=y1 to y2 do
  begin
   zerocount:=0;
   WriteMem32w(ptr48(Selector,offset+(y-y1)*2),idx);
   LineCountIdx:=idx;
   WriteMem32b(ptr48(Selector,offset+LineCountIdx),0);
   inc(idx);            {Save room for linecount}
   x:=x1;
   inzeros:=(getpixel(x,y)=TransParentColour);  {Should we start in zeros?}
   while x<=x2 do           {Continue until we are done with the line}
    case inzeros of
     true:begin
           zerocount:=0;   {How many zeros?}
           while (x<=x2) and (getpixel(x,y)=TransparentColour) do
            begin
             inc(x);
             inc(zerocount,2)
            end;
           inzeros:=FALSE  {we are no longer in zeros}
          end;
     false:begin
            incmem32b(ptr48(Selector,Offset+LineCountIdx),1);
            WriteMem32w(ptr48(Selector,Offset+Idx),ZeroCount);
            inc(idx,2);
            xlcountIdx:=idx; {past zerocount}
            WriteMem32w(ptr48(Selector,Offset+XlCountIdx),0);
            inc(idx,2); {past zerocount and xlcount}
            repeat
             WriteMem32w(ptr48(Selector,Offset+idx),getpixel(x,y));
             inc(x);
             inc(idx,2);
             IncMem32w(ptr48(Selector,offset+xlCountIdx),1);
            until (x>x2) or (getpixel(x,y)=TransparentColour);
            inzeros:=TRUE
           end
   end
  end;
end;

function TGfx16.GetMem;
var tempcount:word;
begin
 GetMem:=FALSE;
 if (CurDataState=OWNER) then exit;
 if (_data=Sega000) then exit;
 if myh>MaxScanLines-1 then exit;
 for tempcount:=0 to 100 do
 begin
  getmem32(_data,myw*longint(myh)*2);
  if _data<>0 then break;
 end;
 if _data=0 then exit;
 CurDataState:=OWNER;
 xRes:=myw*2;
 yRes:=myh;
 fillword32(ptr48(_data,0),myw*longint(myh),0);
 GetMaxX:=myw-1;
 GetMaxY:=myh-1;
 LineOffsets[0]:=0;
 for TempCount:=1 to high(LineOffsets) do
  LineOffsets[TempCount]:=LineOffsets[TempCount-1]+xRes;
 BPP:=16;
 getmem:=TRUE
end;


procedure TGfx16.Flip;
var count,xCount,yCount:word;
begin
 if SrcObject._data=0 then exit;
 if SrcObject.BPP<>16 then exit;
 xCount:=SrcObject.GetMaxX+1;
 if xCount>GetMaxX+1 then xCount:=GetMaxX+1;
 yCount:=SrcObject.GetMaxY;
 if yCount>GetMaxY then yCount:=GetMaxY;
 if SrcObject._data=Sega000 then {This is a vesa16 object, so we need}
  begin                          {to typecast the gfx16 object to a vesa8 obj}
  {(@SrcObject)^ is a dereferenced pointer to the object.  It didn't
   work any other way }
   for count:=0 to yCount do
    TScreen16((@SrcObject)^).CopyLineFromScreen(0,count,
               ptr48(_data,_internalOffset+LineOffsets[count]),xCount)
  end
 else
 for count:=0 to yCount do
  move32(ptr48(SrcObject._data,SrcObject._InternalOffset+SrcObject.LineOffsets[count]),
         ptr48(_data,_internalOffset+LineOffsets[count]),
         xCount*2)
end;

procedure TGfx16.CopyBuf;
var count,xCount,yCount:word;
begin
 if SrcObject._Data=0 then exit;
 if _data=0 then exit;

 xCount:=(x2-x1+1);
 yCount:=(y2-y1+1);

 if (DestX+xCount>GetMaxX+1) then xCount:=GetMaxX-DestX+1;

 if (DestY+yCount>GetMaxY+1) then yCount:=GetMaxY-DestY+1;

 if (DestX<0) then
  begin
   inc(xcount,DestX);
   dec(x1,DestX);
   DestX:=0;
  end;

 if (DestY<0) then
  begin
   inc(yCount,DestY);
   dec(y1,DestY);
   DestY:=0;
  end;

 if (DestX+xCount>GetMaxX+1) or (DestX+xCount<0) or (DestY+yCount>GetMaxY+1) or
    (DestY+yCount<0) then exit;

 if SrcObject._data=Sega000 then
  for count:=0 to yCount-1 do
   Tscreen16((@SrcObject)^).CopyLineFromScreen(x1,y1+count,
         ptr48(_data,_InternalOffset+LineOffsets[DestY+count]+(DestX*2)),
         xCount)
 else
 for count:=0 to yCount-1 do
  move32(ptr48(SrcObject._data,SrcObject._InternalOffset+SrcObject.LineOffsets[y1+count]+(x1*2)),
         ptr48(_data,_internalOffset+LineOffsets[DestY+count]+(DestX*2)),
         xcount*2)
end;



{$L ObjGfx16} {objgfx16.asm}
procedure TGfx16.PutPixel; external;
function TGfx16.GetPixel; external;
procedure TGfx16.Clear; external;
procedure TGfx16.GetImage; external;
procedure TGfx16.PutImage; external;
procedure TGfx16.PutBlit; external;
procedure TGfx16.ScaleBuf; external;
procedure TGfx16.FlipY; external;
procedure TGfx16.horizline; external;
procedure TGfx16.vertline; external;
procedure TGfx16.rawline; external;

procedure TScreen16.PutPixel; external;
function TScreen16.GetPixel; external;
procedure TScreen16.Clear; external;
procedure TScreen16.GetImage; external;
procedure TScreen16.PutImage; external;
procedure TScreen16.CopyLineToScreen; external;
procedure TScreen16.CopyLineFromScreen; external;
procedure TScreen16.horizline; external;
procedure TScreen16.vertline; external;
procedure TScreen16.rawline; external;
procedure TScreen16.PutBlit; external;
procedure TScreen16.ScaleBuf; external;

function RGB(red,green,blue:byte):longint; external;


constructor TScreen16.init;
var MR,VR:word;
begin
 inherited init;
 _VESASelector:=AllocSelector(0);
 _CodeSelector:=0;
 MR:=word(GlobalDosAlloc(sizeof(TMode_Rec)));
 VR:=word(GlobalDosAlloc(sizeof(TVesa_Rec)));
 ModeRec:=ptr(MR,0);
 VesaRec:=ptr(VR,0);
 _CurWriteBank:=65535;
 _CurReadBank:=65535;
 _InGraphics:=FALSE;
 BPP:=16;
end;

procedure TScreen16.CopyBuf;
var count,xCount,yCount:word;
begin
 if _LFB then inherited CopyBuf(DestX,DestY,SrcObject,x1,y1,x2,y2) else
  begin
   if SrcObject._data=0 then exit;
   if _data=0 then exit;
   if SrcObject._data=Sega000 then exit; {do special case later}

   xCount:=(x2-x1+1);
   yCount:=(y2-y1+1);

   if (DestX+xCount>GetMaxX+1) then xCount:=GetMaxX-DestX+1;

   if (DestY+yCount>GetMaxY+1) then yCount:=GetMaxY-DestY+1;

   if (DestX<0) then
    begin
     inc(xcount,DestX);
     dec(x1,DestX);
     DestX:=0;
    end;

   if (DestY<0) then
    begin
     inc(yCount,DestY);
     dec(y1,DestY);
     DestY:=0;
    end;

   With SrcObject do
   for count:=0 to yCount-1 do
    CopyLineToScreen(ptr48(_data,_InternalOffset+LineOffsets[y1+Count]+(x1*2)),
                     DestX,DestY+Count,xCount);
  end;
end;

function TScreen16.GetModeInfo;
begin
 with RMRegs do
  begin
   ax:=$4f01;
   cx:=mode;
   di:=0;
   es:=GetSegBaseAddr(seg(ModeRec^)) shr 4;
   sp:=0; {Clear stack so DPMI handles it}
   ss:=0  {"}
  end;
 RealModeInt($10,RMRegs);
 GetModeInfo:=RMRegs.ax;
end;

function TScreen16.IsModeAvailable;
begin
 if (modeNum>=$100) then
  begin
   GetModeInfo(ModeNum);
   IsModeAvailable:=(ModeRec^.ModeAttributes and 1=1); {yeah, right}
  end else IsModeAvailable:=FALSE;
end;

function TScreen16.IsVesaInstalled;
begin

 with VesaRec^ do
  begin
   VBESignature[0]:='V';
   VBESignature[1]:='B';
   VBESignature[2]:='E';
   VBESignature[3]:='2';
  end;

 with RMRegs do
  begin
   ax:=$4f00;
   di:=0;
   es:=GetSegBaseAddr(seg(VESARec^)) shr 4;
   ss:=0; {Clear stack so DPMI handles it}
   sp:=0  {"}
  end;
 RealModeInt($10,RMRegs);

 {Convert all real mode pointers to protected mode pointers}
 with VesaRec^ do
  begin
   VideoModePtr:=ConvertPtr(VideoModePtr);
   OEMStringPtr:=ConvertPtr(OEMStringPtr);
   OEMVendorNamePtr:=ConvertPtr(OEMVendorNamePtr);
   OEMProductRevPtr:=ConvertPtr(OEMProductRevPtr);
  end;
 IsVesaInstalled:=RMRegs.ax
end;

function TScreen16.UsingLFB:boolean;
begin
 UsingLFB:=_LFB
end;

procedure dummy; far; assembler;
asm end;

procedure OldSwitchbank; far; assembler;
asm
        push    ax
        mov     ax,4f05h
        int     10h
        pop     ax
end;

procedure TScreen16.SetupMode(Mode:word);
var count,GranInc,Rights:word;
    PhysPtr,Size:longint;
    s:string[35];
begin
 {The IsVesaInstalled() call is possibly redundant, but necessary.  If
  someone calls SetupMode() without first calling IsVesaInstalled it will
  blow up.  Because SetMode checks the result of IsVesaInstalled, it is
  okay to throw away the result of it here.  Either they wanted error
  checking (in which case they called SetMode) or they don't want/need
  error checking and called this routine directly.}
 IsVesaInstalled;

 _Data:=Sega000;
 CurDataState:=ALIASING;
 _InGraphics:=TRUE;
 _LFB:=FALSE;

 if not(ForceBanked) and (VesaRec^.MajVersion>=2) then
  begin
   mode:=Mode or $4000; {attempt LFB}
   GetModeInfo(mode);
   with ModeRec^ do
    if physbaseptr=0 then Mode:=Mode and $3FFF  {No LFB, go banked}
     else
      begin
       size:=Xres*Longint(BytesPerLine);
       _LFB:=TRUE;
       PhysPtr:=PhysicalAddressMapping(PhysBasePtr,size);
       _data:=CreateDescriptor(PhysPtr,size);
       rights:=GetSegmentAccessRights(_Data);
       rights:=(rights and $FF70) or $0093;
       SetSegmentAccessRights(_data,rights)
      end
  end;

 GetModeInfo(mode);

 Xres:=ModeRec^.BytesPerLine;
 YRes:=ModeRec^.YRes;
 GetMaxX:=ModeRec^.Xres-1;
 GetMaxY:=YRes-1;

 LineOffsets[0]:=0;
 for count:=1 to high(LineOffsets) do
  LineOffsets[count]:=LineOffsets[count-1]+ModeRec^.BytesPerLine;

 GranInc:=1;
 if ModeRec^.Granularity<>0 then
  GranInc:=64 div ModeRec^.Granularity; {Adjust for page size}
  _BankVals[0]:=0;
 for count:=1 to high(_bankvals) do
  _BankVals[count]:=_BankVals[count-1]+GranInc;


 RedFieldPosition:=ModeRec^.RedFieldPosition;
 GreenFieldPosition:=ModeRec^.GreenFieldPosition;
 BlueFieldPosition:=ModeRec^.BlueFieldPosition;

 RedShifter:=8-ModeRec^.RedMaskSize;
 GreenShifter:=8-ModeRec^.GreenMaskSize;
 BlueShifter:=8-ModeRec^.BlueMaskSize;

 count:=0;
 with vesarec^ do
 while (count<32) and (pbytearray(OEMStringPtr)^[count]<>0) do
  begin
   s[count+1]:=upcase(char(pbytearray(OEMStringPtr)^[count]));
   inc(count);
  end;
 s[0]:=chr(count+1);
 if (pos('UNIVERSAL',s)>0) or (ForceInt10h) or (ModeRec^.BankSwitch=NIL)
  then @SwitchBank:=@OldSwitchBank
 else
  begin
   SetSelectorBase(_VESASelector,longint(dword(ModeRec^.BankSwitch).HiWord)*16);
   SetSelectorLimit(_VESASelector,65535);
   if _codeselector<>0 then FreeSelector(_CodeSelector);
   _CodeSelector:=AllocDStoCSAlias(_VesaSelector);
   @SwitchBank:=ptr(_CodeSelector,Dword(ModeRec^.BankSwitch).LoWord);
  end;
 If ModeRec^.WindowAFlags and 2=2 then
  _ReadWindow:=0
 else if ModeRec^.WindowBFlags and 2=2 then
  _ReadWindow:=1; {else there is no readable window!?}
 InitVESAMode(mode);
 _CurReadBank:=0;
 _CurWriteBank:=0;
end;

function TScreen16.setmode;
begin
 SetMode:=FALSE; {presume failure}
 if IsVesaInstalled<>VESA_ok then exit;
 if not IsModeAvailable(mode) then exit;
 setupMode(mode);
 SetMode:=TRUE
end;

procedure TScreen16.Flip;
var count,xCount,yCount:word;
begin
 if SrcObject._data=0 then exit;
 if SrcObject._data=SegA000 then exit; {This would be flipping ourselves}
 if _LFB then inherited Flip(SrcObject) else
  begin
   xCount:=SrcObject.GetMaxX+1;
   if xCount>GetMaxX+1 then xCount:=GetMaxX+1;
   yCount:=SrcObject.GetMaxY;
   if yCount>GetMaxY then yCount:=GetMaxY;
   With SrcObject Do
   for count:=0 to yCount do
    CopyLineToScreen(ptr48(_data,_InternalOffset+LineOffsets[count]),
                     0,count,xCount);
  end;
end;

procedure TScreen16.WaitForRetrace; assembler; {Wait for Retrace}
asm
	mov     dx,03dah
@gn0:   in      al,dx
	test    al,8
	jnz     @gn0

@gn1:   in      al,dx
	test    al,8
	jz      @gn1
end;

destructor TScreen16.done;
begin
 inherited done;
 if _InGraphics then asm mov ax,3; int 10h end;
 if _VESASelector<>0 then FreeSelector(_VESASelector);
 if _CodeSelector<>0 then FreeSelector(_CodeSelector);
 if VesaRec<>NIL then globaldosfree(seg(VesaRec^));
 if ModeRec<>NIL then globaldosfree(seg(ModeRec^))
end;

procedure GetVidMode; assembler;
asm
        mov     ax,4f03h
	int     $10
	mov     OldVidMode,bx
end;

begin
 RedShifter         :=3;
 GreenShifter       :=2;
 BlueShifter        :=3;
 RedFieldPosition   :=11;
 GreenFieldPosition :=5;
 BlueFieldPosition  :=0;
end.