(********************************************************************)
(*                                                                  *)
(*                               OBJGFX8.PAS                        *)
(*                                                                  *)
(*                     Object Oriented Graphics Library             *)
(*                                                                  *)
(*                                                                  *)
(*                            Copyright (c) 1999                    *)
(*                      Mark J Iuzzolino, Daniel Robbins            *)
(*                            All Rights Reserved                   *)
(*                                                                  *)
(*                                                                  *)
(*                                                                  *)
(********************************************************************)

(*
 *   Rev 2.0  1999/05/24 mji
 * Split ObjGfx into ObjGfx0 and ObjGfx8
 *
 *   Rev 1.9  1999/05/09 mji
 * Removed all font related routines
 *
 *   Rev 1.8  1998/09/28 mji
 * Released to the general public
 *
 *   Rev 1.0  1998/03/10 mji
 * Version 1.0 released to the general public
 *)

(*
 * Presumed data types and sizes (for all you non-pascal programmers):
 *
 * boolean   : 8-bit boolean value
 * wordbool  : 16-bit boolean value
 * byte      : unsigned 8-bit value
 * shortint  : signed 8-bit value
 * char      : unsigned 8-bit value (character)
 * word      : unsigned 16-bit value
 * integer   : signed 16-bit value
 * longint   : signed 32-bit value
 * real      : 48-bit floating point
 * pointer   : 32-bit pointer (16-bit selector; 16-bit offset)
 * pointer48 : 48-bit pointer (16-bit selector; 32-bit offset)
 * string    : array of char; string length is string[0]
 *
 *)

{$A+,B-,D+,E+,F-,G+,I+,L+,N-,P-,Q-,R-,S+,T-,V+,X+,Y+}
Unit ObjGfx8;

interface

uses ObjGfx0,NewFrontier;

const _320x200x8bpp       = $13;
const _640x400x8bpp       = $100;
const _640x480x8bpp       = $101;
const _800x600x8bpp       = $103;
const _1024x768x8bpp      = $105;
const _1280x1024x8bpp     = $107;

const VESA_Ok             = $4F;

type PGfx8 = ^TGfx8;
     TGfx8 = object(TGfx0)
       pal                : PalType;
       procedure AliasGfx8(var Gfx8Object:TGfx8;x1,y1,x2,y2:word); virtual;
       procedure Clear(colour:longint); virtual;
       procedure copybuf(DestX,DestY:integer;var SrcObject:Tgfx0;x1,y1,x2,y2:word); virtual;
       procedure Flip(var SrcObject:TGfx0); virtual;
       procedure FlipX; virtual;
       procedure FlipY; virtual;
       procedure horizline(x1,x2,y:integer;colour:longint); virtual;
       procedure getblit(const x1,y1,x2,y2:word;blit:pointer48); virtual;
       function GetMem(myw,myh:word):boolean;
       procedure GetImage(x1,y1,x2,y2:word;bitmap:pointer48); virtual;
       function GetPixel(x,y:integer):longint; virtual;
       procedure GetRGBPalette(colour:byte;var red,green,blue:byte); virtual;
       function LoadGfx(const gfxfile:string):boolean;
       procedure LoadPal(const palfile:string);
       procedure putblit(x,y:Integer;blit:pointer48); virtual;
       procedure putimage(x,y:integer;bitmap:pointer48); virtual;
       procedure PutPixel(x,y:integer;colour:longint); virtual;
       procedure RawLine(x1,y1,x2,y2:word;colour:longint); virtual;
       function SaveGfx(const gfxfile:string;GfxType:TGfxType):boolean;
       procedure Savepal(const palfile:string);
       procedure scalebuf(dX1,dY1,dX2,dY2:integer;var SrcObject:Tgfx0;sx1,sy1,sx2,sy2:word); virtual;
       procedure SetRGBPalette(colour,red,green,blue:byte); virtual;
       procedure StripPal(const palfile:string);
       procedure vertline(x,y1,y2:integer;colour:longint); virtual;
     end;

     PScreen8 = ^TScreen8;
     TScreen8 = object(TGfx8)
       VESARec            : ^TVESA_Rec;
       ModeRec            : ^TMode_Rec;
       SwitchBank         : procedure;
       _BankVals          : array[0..479] of word;
       _CurWriteBank      : word;
       _CurReadBank       : word;
       _ReadWindow        : word;
       _VESASelector      : word;
       _CodeSelector      : word;
       _InGraphics        : boolean;
       _LFB               : boolean;
       constructor init;
       procedure Clear(colour:longint); virtual;
       procedure copybuf(DestX,DestY:integer;var SrcObject:Tgfx0;x1,y1,x2,y2:word); virtual;
       procedure CopyLineFromScreen(x,y:word;src:pointer48;size:word);
       procedure CopyLineToScreen(src:pointer48;x,y:word;size:word);
       procedure GetImage(x1,y1,x2,y2:word;bitmap:pointer48); virtual;
       function GetModeInfo(mode:word):word;
       procedure GetPal;
       function GetPixel(x,y:integer):longint; virtual;
       procedure horizline(x1,x2,y:integer;colour:longint); virtual;
       function IsModeAvailable(ModeNum:word):boolean;
       function IsVesaInstalled:word;
       procedure flip(var SrcObject:TGfx0); virtual;
       procedure MakePal;
       procedure putblit(x,y:Integer;blit:pointer48); virtual;
       procedure putimage(x,y:integer;bitmap:pointer48); virtual;
       procedure PutPixel(x,y:integer;colour:longint); virtual;
       procedure RawLine(x1,y1,x2,y2:word;colour:longint); virtual;
       procedure scalebuf(dX1,dY1,dX2,dY2:integer;var SrcObject:TGfx0;sx1,sy1,sx2,sy2:word); virtual;
       function SetMode(mode:word):boolean;
       procedure SetRGBPalette(colour,red,green,blue:byte); virtual;
       procedure SetupMode(mode:word);
       function UsingLFB:boolean;
       procedure vertline(x,y1,y2:integer;colour:longint); virtual;
       procedure WaitForRetrace;
       destructor done; virtual;
     end;

function WhatType8(buffer:pointer):TGfxType;
procedure initVESAmode(mode:word);

implementation

uses dos,winapi;

{$l objgfx8} {objgfx8.asm}
{$F+}
procedure CompressLine_Mooishi(src,dest:pointer48;EndOfLine:longint;
                               compclr1,compclr2,comprclr3,compclr4:byte;
                               var size:word); external;
procedure DeCompressLine_Mooishi(src,dest:pointer48;EndOfLine:longint;
                               compclr1,compclr2,comprclr3,compclr4:byte;
                               var size:word); external;

procedure CompressLine_PCX(src,dest:pointer48;EndOfLine:longint;var size:word); external;
procedure DeCompressLine_PCX(src,dest:pointer48;EndOfLine:longint;var size:word); external;
procedure DeCompressLine_PBM(src,dest:pointer48;EndOfLine:longint;var size:word); external;
procedure DeCompressLine_IFF(src,dest:pointer48;EndOfLine:longint;var size:word); external;
{$F-}
procedure initVESAmode(mode:word); assembler;
asm
        mov     ax,4f02h
        mov     bx,mode
        int     10h
end;

function WhatType8(buffer:pointer):TGfxType;
begin
 with PNGSignature(buffer^),
      MooishiHeader(buffer^),
      WinBmpFileHeader(buffer^),
      GifHeader(buffer^),
      ilbmchunk(buffer^),
      PCXHeader(buffer^),
      Win3xBitMapHeader(buffer^) do
  {Detect if it is a PNG file type}
  if ((bswapl(Sig1)=$89504e47) and (bswapl(Sig2)=$0D0A1A0A)) then WhatType8:=PNG
  else {Not a PNG, check if it is a Windows BMP}
  if (FileType=$4d42) and (BitsPerPixel=8) and (planes=1) then
   begin
    if Compression=0 then WhatType8:=BMP else
      if Compression=1 then WhatType8:=COMPBMP else WhatType8:=NOGRAPHIC;
   end
  else {Not a PNG or BMP, check if it is a MSH}
   if ((ID_String[0]='M') and (ID_String[1]='O') and (ID_String[2]='O') and
      (ID_String[3]='I') and (ID_String[4]='S') and (ID_String[5]='H') and
      (ID_String[6]='I') and (ID_String[7]='!')) then WhatType8:=MSH
   else {Not a PNG, BMP, or MSH, check if it is a GIF}
    if ((Signature[0]='G') and (Signature[1]='I') and (Signature[2]='F')) then WhatType8:=GIF
    else {Not a PNG, BMP, MSH, or GIF, check if it is an IFF}
     if ((ChunkID[0]='F') and (ChunkID[1]='O') and (ChunkID[2]='R') and
        (ChunkID[3]='M')) then
         begin
          WhatType8:=LBM;
          if tbytearray(buffer^)[8]=ord('I') then WhatType8:=ILBM
           else if tbytearray(buffer^)[8]=ord('P') then WhatType8:=LBM
            else WhatType8:=NOGRAPHIC;
         end
     else {Not a PNG, BMP, MSH, GIF, or IFF, check if it is a PCX}
      if (Manufacturer=$0A) and (bits_per_pixel=8) and (colour_planes=1) then WhatType8:=PCX
      else WhatType8:=NOGRAPHIC; {It's not something I can detect}
end;

procedure TGfx8.AliasGfx8;
var count,width,height:word;
begin
 if CurDataState<>OWNER then
  begin
   if x2<x1 then
    begin
     count:=x2;
     x2:=x1;
     x1:=count
    end;
   if y2<y1 then
    begin
     count:=y2;
     y2:=y1;
     y1:=count
    end;
   width:=(x2-x1)+1;
   height:=(y2-y1)+1;
   if height>MaxScanLines then height:=MaxScanLines;
   CurDataState:=ALIASING;
   Xres:=Gfx8Object.Xres;
   YRes:=Gfx8Object.YRes;
   GetMaxX:=width-1;
   GetMaxY:=height-1;
   _data:=Gfx8Object._Data;
   LineOffsets[0]:=0;
   for count:=1 to height-1 do
    LineOffsets[Count]:=LineOffsets[Count-1]+Xres;
   _InternalOffset:=Gfx8Object.LineOffsets[y1]+x1+Gfx8Object._InternalOffset;
  end
end;

procedure TGfx8.Clear; external;

procedure TGfx8.CopyBuf;
{(DestX,DestY:integer;var SrcObject:gfx8;x1,y1,x2,y2:word);}
var count,xCount,yCount:integer;
begin
 if SrcObject._Data=0 then exit;
 if _Data=0 then exit; {duh, check to see if we can actually draw it!}

 xCount:=(x2-x1+1);
 yCount:=(y2-y1+1);

 if (DestX+xCount>GetMaxX+1) then xCount:=GetMaxX-DestX+1;

 if (DestY+yCount>GetMaxY+1) then yCount:=GetMaxY-DestY+1;

 if (DestX<0) then
  begin
   inc(xcount,DestX);
   dec(x1,DestX);
   DestX:=0;
  end;

 if (DestY<0) then
  begin
   inc(yCount,DestY);
   dec(y1,DestY);
   DestY:=0;
  end;

 if (DestX+xCount>GetMaxX+1) or (DestX+xCount<0) or (DestY+yCount>GetMaxY+1) or
    (DestY+yCount<0) then exit;

 if SrcObject._data=Sega000 then
  for count:=0 to yCount-1 do
   TScreen8((@SrcObject)^).CopyLineFromScreen(x1,y1+count,
         ptr48(_data,_InternalOffset+LineOffsets[DestY+count]+DestX),
         xCount)
 else
  for count:=0 to yCount-1 do
   move32(ptr48(SrcObject._data,SrcObject._InternalOffset+SrcObject.LineOffsets[y1+count]+x1),
          ptr48(_data,_internalOffset+LineOffsets[DestY+Count]+DestX),
          xcount);
end;

procedure TGfx8.Flip;
var count,xCount,yCount:word;
begin
 if SrcObject._data=0 then exit;
 xCount:=SrcObject.GetMaxX+1;
 if xCount>GetMaxX+1 then xCount:=GetMaxX+1;
 yCount:=SrcObject.GetMaxY;
 if yCount>GetMaxY then yCount:=GetMaxY;
 if SrcObject._data=Sega000 then {This is a vesa8 object, so we need}
  begin                          {to typecast the gfx8 object to a vesa8 obj}
  {(@SrcObject)^ is a dereferenced pointer to the object.  It didn't
   work any other way }
   for count:=0 to yCount do
    TScreen8((@SrcObject)^).CopyLineFromScreen(0,count,
               ptr48(_data,_internalOffset+LineOffsets[count]),xCount)
  end
 else
 for count:=0 to yCount do
  move32(ptr48(SrcObject._data,SrcObject._InternalOffset+SrcObject.LineOffsets[count]),
         ptr48(_data,_internalOffset+LineOffsets[count]),
         xCount)
end;

procedure TGfx8.FlipX;
var xwidth,count,buf:word;
begin
 xwidth:=GetMaxX+1;
 getmem32(buf,XWidth);
 for count:=0 to (GetMaxY div 2) do
  begin
   move32(ptr48(_data,_InternalOffset+LineOffsets[count]),ptr48(buf,0),Xwidth);
   move32(ptr48(_data,_InternalOffset+LineOffsets[GetMaxY-count]),
          ptr48(_data,_InternalOffset+LineOffsets[count]),xwidth);
   move32(ptr48(buf,0),ptr48(_data,_InternalOffset+LineOffsets[GetMaxY-count]),xwidth);
  end;
 freemem32(buf)
end;

procedure TGfx8.FlipY; external;

function TGfx8.GetMem;
var tempcount:word;
begin
 GetMem:=FALSE;
 if (CurDataState=OWNER) then exit;
 if (_data=Sega000) then exit;
 if myh>MaxScanLines then exit;{max height is 4096 pixels; width is limited to 32768}
 getmem32(_data,myw*longint(myh));
 if _data=0 then exit;
 CurDataState:=OWNER;
 fillchar32(ptr48(_data,0),myw*longint(myh),0);
 BPP:=8;
 xRes:=myw;
 yRes:=myh;
 GetMaxX:=myw-1;
 GetMaxY:=myh-1;
 LineOffsets[0]:=0;
 for TempCount:=1 to high(LineOffsets) do
  LineOffsets[TempCount]:=LineOffsets[TempCount-1]+xRes;
 getmem:=TRUE;
end;

procedure TGfx8.getblit;
var offset:longint;
    selector,x,y,idx:word;
    zerocount:word;
    found,InZeros:boolean;
    LineCountIdx,xlcountIdx:word;
    StartX,StartY,EndX,Endy:word; {relative pic coords}
begin
 Found:=FALSE;
 StartX:=x1;
 StartY:=y1;
 EndX:=x2;
 EndY:=y2;
 while not(found) and (StartX<x2) do
  begin
   for y:=y1 to y2 do
    found:=found or (getpixel(StartX,y)<>TransparentColour);
   if not found then inc(StartX)
  end;
 found:=FALSE;
 while not(found) and (StartY<y2) do
  begin
   for x:=StartX to x2 do
    found:=found or (getpixel(x,StartY)<>TransparentColour);
   if not found then inc(StartY);
  end;
 found:=FALSE;
 while not(found) and (EndX>StartX) do
  begin
   for y:=StartY to y2 do
    found:=found or (getpixel(EndX,y)<>TransparentColour);
   if not found then dec(EndX);
  end;
 found:=FALSE;
 while not(found) and (EndY>StartY) do
  begin
   for x:=StartX to EndX do
    found:=found or (getpixel(x,EndY)<>TransparentColour);
   if not found then dec(EndY);
  end;
 selector:=seg48(blit);
 offset:=ofs48(blit);
 WriteMem32w(ptr48(selector,offset),StartX-x1);
 WriteMem32w(ptr48(selector,offset+2),StartY-y1);
 WriteMem32w(ptr48(selector,offset+4),EndX-x1);
 WriteMem32w(ptr48(selector,offset+6),EndY-y1);
 inc(offset,8);
 idx:=((y2-y1)+1)*2;
 for y:=y1 to y2 do
  begin
   zerocount:=0;
   WriteMem32w(ptr48(selector,offset+(y-y1)*2),idx);
   LineCountIdx:=idx;
   WriteMem32b(ptr48(selector,offset+LineCountIdx),0);
   inc(idx);            {Save room for linecount}
   x:=x1;
   inzeros:=(getpixel(x,y)=TransParentColour);  {Should we start in zeros?}
   while x<=x2 do           {Continue until we are done with the line}
    case inzeros of
     true:begin
           zerocount:=0;   {How many zeros?}
           while (x<=x2) and (getpixel(x,y)=TransparentColour) do
            begin
             inc(x);
             inc(zerocount)
            end;
           inzeros:=FALSE  {we are no longer in zeros}
          end;
     false:begin
            incmem32b(ptr48(selector,Offset+LineCountIdx),1);
            WriteMem32w(ptr48(selector,Offset+Idx),ZeroCount);
            inc(idx,2);
            xlcountIdx:=idx; {past zerocount}
            WriteMem32w(ptr48(selector,Offset+XlCountIdx),0);
            inc(idx,2); {past zerocount and xlcount}
            repeat
             WriteMem32b(ptr48(selector,offset+idx),getpixel(x,y));
             inc(x);
             inc(idx);
             IncMem32w(ptr48(selector,offset+xlCountIdx),1);
            until (x>x2) or (getpixel(x,y)=TransparentColour);
            inzeros:=TRUE
           end
   end
  end;
end;

procedure TGfx8.GetImage; external;
function TGfx8.GetPixel; external;

procedure TGfx8.GetRGBPalette;
begin
 with pal[colour] do
  begin
   red:=r;
   green:=g;
   blue:=b
  end
end;

procedure TGfx8.HorizLine; external;

function TGfx8.LoadGfx;
var infile:file;
    C_Bufsize:longint;
    info:pointer;
    CompressedBuf:word;
    tmpbuf:word;
    GfxType:TGfxType;
    srcidx,destidx,tmpidx:longint;
    ddy,resultsize:word;
    LinePaddingSize:word;
    myh,myw:word;
    bool:boolean;
label safeexit; {the dreaded goto!!}
begin
 LoadGfx:=FALSE;
 if fsearch(gfxfile,'')='' then exit;
 system.getmem(Info,128);
 assign(infile,gfxfile);
 reset(infile,1);
 C_BufSize:=filesize(InFile);
 blockread(infile,Info^,128);
 close(infile);
 if C_BufSize=64768 then GfxType:=DNY else GfxType:=WhatType8(info);
 if (GfxType=NOGRAPHIC) then goto safeexit;
 GetMem32(CompressedBuf,C_BufSize);
 if (CompressedBuf=0) then goto safeexit;
 bload32(gfxfile,ptr48(CompressedBuf,0),C_BufSize);
 case GfxType of
  PCX:with PCXHeader(info^) do
       begin
        if (CurdataState=NONE) and not(getmem(Xmax-Xmin+1,Ymax-Ymin+1))
         then goto safeexit;
        srcidx:=128;
        for ddy:=0 to (YMax-Ymin) do
         begin
          destidx:=_internalOffset+LineOffsets[ddy];
          deCompressLine_PCX(ptr48(CompressedBuf,srcidx),
                             ptr48(_data,destidx),
                             destidx+xmax-xmin+1,
                             ResultSize);
          inc(srcidx,ResultSize);
         end;
       end;
  MSH:with MooishiHeader(info^) do
       begin
        if (CurDataState=NONE) and not(getmem(hres,vres)) then goto safeexit;
        if (hres=320) and (vres=200) then
         begin
          getmem32(tmpbuf,64000);
          if tmpbuf=0 then goto safeexit;
          decompressline_mooishi(ptr48(compressedbuf,598),ptr48(tmpbuf,0),64000,
                                 Escbytes[0],escbytes[1],escbytes[2],escbytes[3],ResultSize);
          for ddy:=0 to 199 do
           move32(ptr48(tmpbuf,ddy*320),ptr48(_data,_internaloffset+LineOffsets[ddy]),320);
          freemem32(tmpbuf)
         end
        else
        begin
         srcidx:=598;
         for ddy:=0 to vres-1 do
          begin
           destidx:=_internaloffset+LineOffsets[ddy];
           deCompressLine_Mooishi(ptr48(CompressedBuf,srcidx),
                                  ptr48(_data,destidx),
                                  destidx+hres,
                                  Escbytes[0],escbytes[1],escbytes[2],escbytes[3],
                                  ResultSize);
           inc(srcidx,ResultSize)
          end
        end
       end;
  DNY:begin
       if (CurDataState=NONE) and not(getmem(320,200)) then goto safeexit;
       for ddy:=0 to 199 do
        move32(ptr48(CompressedBuf,ddy*320+768),
               ptr48(_data,_InternalOffset+LineOffsets[ddy]),
               320)
      end;
  BMP:with Win3xBitmapHeader(info^) do
       begin
        bool:=(height<0);
        height:=abs(height);
        if (CurDataState=NONE) and not(getmem(width,height)) then goto safeexit;
        LinePaddingSize:=(4-(width and 3)) and 3;
        srcidx:=FileHeader.BitmapOffset;
        for ddy:=0 to height-1 do
         begin
          move32(ptr48(CompressedBuf,srcidx+ddy*width),
                 ptr48(_data,_InternalOffset+LineOffsets[GetMaxY-ddy]),
                 width);
          inc(srcidx,LinePaddingSize);
         end;
        if bool then flipX
       end;
  ILBM,LBM:begin
       bool:=FALSE;
       tmpidx:=0;
       srcidx:=12;
       repeat
        move32(ptr48(CompressedBuf,srcidx),farptr(info),sizeof(ILBMChunk));
        inc(srcidx,sizeof(ILBMChunk));
        with ILBMChunk(info^) do
         begin
          bool:=(ChunkID[0]='B') and (ChunkID[1]='M') and
                (ChunkID[2]='H') and (ChunkID[3]='D');
          size:=bswapl(size);
          if odd(size) then inc(size);
          if not bool then inc(srcidx,size);
         end;
       until (bool) or (srcidx>=C_bufsize);
       if not bool then goto safeexit;
       move32(ptr48(CompressedBuf,srcidx),farptr(info),sizeof(BitMapHeaDer));
       inc(srcidx,sizeof(BitMapHeaDer));
       with BitMapHeaDer(info^) do
        begin
         width:=bswapw(width);
         height:=bswapw(height);
         myw:=width;
         myh:=height;
         if (CurDataState=NONE) and not(getmem(width,height)) then goto safeexit;
        end;
       bool:=FALSE;
       {look for the BODY chunk}
       repeat
        move32(ptr48(CompressedBuf,srcidx),farptr(info),sizeof(ILBMChunk));
        inc(srcidx,sizeof(ILBMChunk));
        with ILBMChunk(info^) do
         begin
          bool:=(ChunkID[0]='B') and (ChunkID[1]='O') and
                (ChunkID[2]='D') and (ChunkID[3]='Y');
          size:=bswapl(size);
          if odd(size) then inc(size);
          if not bool then inc(srcidx,size);
         end;
       until (bool) or (srcidx>=C_bufsize);
       if not bool then goto safeexit;
       if GfxType=LBM then
       for ddy:=0 to myh-1 do
        begin
         destidx:=_internalOffset+LineOffsets[ddy];
         deCompressLine_PBM(ptr48(CompressedBuf,srcidx),
                            ptr48(_data,destidx),
                            destidx+myw,
                            ResultSize);
         inc(srcidx,ResultSize)
        end
       else
       for ddy:=0 to myh-1 do
        begin
         destidx:=_internalOffset+LineOffsets[ddy];
         deCompressLine_IFF(ptr48(CompressedBuf,srcidx),
                            ptr48(_data,destidx),
                            destidx+myw,
                            ResultSize);
         inc(srcidx,ResultSize)
        end;
     end;
 end;
 strippal(gfxfile);
 LoadGfx:=TRUE;
safeexit:
 freemem32(CompressedBuf);
 system.freemem(Info,128)
end;

procedure TGfx8.LoadPal;
var f:file;
begin
 assign(f,palfile);
 {$I-} reset(f,1); {$I+}
 blockread(f,pal,sizeof(pal));
 close(f)
end;

procedure TGfx8.PutBlit; external;
procedure TGfx8.PutImage; external;
procedure TGfx8.PutPixel; external;
procedure TGfx8.RawLine; external;

function TGfx8.SaveGfx(const gfxfile:string;GfxType:TGfxType):boolean;
const ID_S:array[0..7] of char='MOOISHI!';
var header:pointer;
    ColourUsage:array[byte] of byte;
    x,y,index,count,tmp,linesize:word;
    cbuf:word;
    colour:byte;
    result:longint;
    plte:Win3xPaletteElement;
    ID_Chunk:array[0..3] of char;
    f:file;
begin
 SaveGfx:=FALSE;
 if (_data=0) or (_data=Sega000) then exit;
 system.getmem(header,256);
 assign(f,gfxfile);
 rewrite(f,1);
 case GfxType of
  DNY:if (GetMaxX=319) and (GetMaxY=199) then
       begin
        blockwrite(f,pal,sizeof(pal));
        for count:=0 to 199 do
         blockwrite32(f,ptr48(_data,_InternalOffset+lineoffsets[count]),320,result);
       end;
  MSH:begin
       fillchar(ColourUsage,sizeof(ColourUsage),255);

       for y:=0 to getmaxy do
        for x:=0 to getmaxx do
         ColourUsage[getpixel(x,y)]:=0;

       index:=0;
       count:=0;
       with MooishiHeader(header^) do
        begin
         EscBytes[0]:=$ff;
         EscBytes[1]:=$fe;
         EscBytes[2]:=$fd;
         EscBytes[3]:=$fc;

         while (index<=$ff) and (count<=3) do
          begin
           if ColourUsage[index]<>0 then
            begin
             EscBytes[count]:=index;
             inc(count)
            end;
           inc(index)
          end;

         move(ID_S,ID_String,sizeof(ID_S));
         hres:=getmaxx+1;
         vres:=getmaxy+1;
         fillchar(reserved,sizeof(reserved),0);
        end;
       blockwrite(f,header^,sizeof(MooishiHeader));
       index:=0;
       repeat
        colour:=RawPalType(pal)[index] shl 2;
        inc(index);
        colour:=colour or (RawPalType(pal)[index] shr 4);
        blockwrite(f,colour,1);

        colour:=RawPalType(pal)[index] shl 4;
        inc(index);
        colour:=colour or (RawPalType(pal)[index] shr 2);
        blockwrite(f,colour,1);

        colour:=RawPalType(pal)[index] shl 6;
        inc(index);
        colour:=colour or RawPalType(pal)[index];
        inc(index);
        blockwrite(f,colour,1);
       until index=768;
       getmem32(cbuf,(getmaxx+1)*2);
       with MooishiHeader(header^) do
        for y:=0 to getmaxy do
         begin
          compressline_Mooishi(ptr48(_data,_InternalOffset+LineOffsets[y]),
                               ptr48(cbuf,0),
                               _InternalOffset+LineOffsets[y]+GetMaxX+1,
                               EscBytes[0],EscBytes[1],EscBytes[2],EscBytes[3],
                               linesize);
         blockwrite32(f,ptr48(cbuf,0),linesize,result);
        end;
       freemem32(cbuf);
      end;
  PCX:begin
       with PCXHeader(header^) do
        begin
         Manufacturer:=10;
         Version:=5;
         Encoding:=1;
         Bits_Per_Pixel:=8;
         Xmin:=0;
         Ymin:=0;
         Xmax:=getmaxx;
         YMax:=GetMaxY;
         hdpi:=GetMaxX+1;
         vdpi:=GetMaxY+1;
         fillchar(palette,sizeof(palette),0);
         reserved:=0;
         Colour_planes:=1;
         Bytes_per_line:=GetMaxX+1;
         HorzScreenSize:=0;
         VertScreenSize:=0;
         if Odd(Bytes_per_Line) then inc(Bytes_Per_Line);
         fillchar(filler,sizeof(filler),0);
        end;
       blockwrite(f,header^,sizeof(PCXHeader));
       getmem32(cbuf,(getmaxx+1)*2);
       colour:=0;
       for y:=0 to getmaxy do
        begin
         compressline_pcx(ptr48(_data,_InternalOffset+LineOffsets[y]),
                          ptr48(cbuf,0),
                          _InternalOffset+LineOffsets[y]+GetMaxX+1,
                          linesize);
         blockwrite32(f,ptr48(cbuf,0),linesize,result);
        end;
        colour:=$c0;
        blockwrite(f,colour,1);
        for y:=0 to 2 do
         begin
          for index:=0 to 255 do
           colourusage[index]:=RawPalType(pal)[index+y*256] shl 2;
          blockwrite(f,colourusage,sizeof(colourusage));
         end;
       freemem32(cbuf);
      end;
   BMP:with Win3xBitmapHeader(header^) do
        begin
         fillchar(ColourUsage,sizeof(ColourUsage),0);

         for y:=0 to GetMaxY do
          for x:=0 to GetMaxX do
           ColourUsage[getpixel(x,y)]:=1;
         colour:=0;

         for x:=0 to high(ColourUsage) do
          inc(colour,ColourUsage[x]);

         Size:=sizeof(Win3xBitmapHeader)-sizeof(FileHeader);
         width:=GetMaxX+1;
         linesize:=((width+3) shr 2) shl 2;
         Height:=(GetMaxY+1);
         with FileHeader do
          begin
           FileType:=$4d42;
           BitMapOffset:=sizeof(Win3xBitmapHeader)+sizeof(Win3xPaletteElement)*256;
           FileSize:=BitMapOffset+LineSize*longint(height);
           Reserved1:=0;
           Reserved2:=0;
          end;
          Planes:=1;
          BitsPerPixel:=8;
          Compression:=0;
          HorzResolution:=0;
          VertResolution:=0;
          ColoursUsed:=1 shl bitsperpixel;
          ColoursImportant:=colour;
          blockwrite(f,header^,sizeof(Win3xBitMapHeader));
          plte.pad:=0;
          for x:=0 to 255 do
           with pal[x], plte do
           begin
            red:=r shl 2;
            green:=g shl 2;
            blue:=b shl 2;
            blockwrite(f,plte,sizeof(Win3xPaletteElement))
           end;
          plte.red:=0;
          plte.green:=0;
          plte.blue:=0;
          tmp:=linesize-width;
          for y:=GetMaxY downto 0 do
           begin
            blockwrite32(f,ptr48(_data,_InternalOffset+LineOffsets[y]),width,result);
            if tmp<>0 then blockwrite(f,plte,tmp)
           end
        end;
 end;
 close(f);
 system.freemem(header,256);
 SaveGfx:=TRUE;
end;

procedure TGfx8.SavePal;
var f:file;
begin
 assign(f,palfile);
 {$I-} rewrite(f,1); {$I+}
 blockwrite(f,pal,sizeof(pal));
 close(f)
end;

procedure TGfx8.ScaleBuf; external;

procedure TGfx8.SetRGBPalette;
begin
 with pal[colour] do
  begin
   r:=red;
   g:=green;
   b:=blue
  end
end;

procedure TGfx8.StripPal;
var f:file;
    fsize:longint;
    result:word;
    buf:pointer;
    count:byte;
    found:boolean;
    plte:Win3xPaletteElement;
    tmpsize:word;
begin
 if fsearch(PalFile,'')='' then exit;
 assign(f,palfile);
 reset(f,1);
 fsize:=filesize(f);
 if (fsize=768) or (fsize=64768) then blockread(f,pal,768)
  else
   begin
    system.getmem(buf,1024);
    blockread(f,buf^,1024,result);
    case WhatType8(buf) of
     MSH:begin
          seek(f,22);
          blockread(f,buf^,576);
          asm
        push    ds
	mov	cx,192
        les     di,self
        mov     ax,offset pal
        add     di,ax
        lds     si,buf
@repeat:lodsb
	shr	al,2
	stosb
	dec	si
	lodsb
	shl	al,4
	and	al,00110000b
	mov	ah,al
	lodsb
	shr	al,4
	add	al,ah
	stosb
	dec	si
	lodsb
	shl	al,2
	and	al,00111100b
	mov	ah,al
	lodsb
	shr	al,6
	add	al,ah
	stosb
	dec	si
	lodsb
	and	al,00111111b
	stosb
        dec     cx
        jnz     @repeat
        pop     ds
        end;
         end;
     PCX:begin
          seek(f,fsize-768);
          blockread(f,pal,768);
          for count:=0 to 255 do with pal[count] do
           begin
            r:=r shr 2;
            g:=g shr 2;
            b:=b shr 2;
           end;
         end;
     PNG:begin
          seek(f,8);
          found:=FALSE;
          blockread(f,buf^,8);
          blockread(f,buf^,sizeof(IHDRChunk));
          if IHDRChunk(buf^).ColourType=3 then
           begin
            seek(f,filepos(f)+4);
            with PNGChunk(buf^) do
            repeat
             blockread(f,buf^,8);
             DataLength:=bswapl(DataLength);
             if ((DataType[0]='P') and (DataType[1]='L') and
                 (DataType[2]='T') and (DataType[3]='E')) then
                  begin
                   Found:=TRUE;
                   blockread(f,pal,DataLength);
                   for count:=0 to (DataLength div 3)-1 do
                    with pal[count] do
                    begin
                     r:=r shr 2;
                     g:=g shr 2;
                     b:=b shr 2
                    end;
                  end
             else seek(f,filepos(f)+DataLength+4);
            until found or ((DataType[0]='I') and
                  (DataType[1]='D') and (DataType[2]='A') and
                  (DataType[3]='T'));
           end
         end;
     ILBM,LBM:begin
          found:=FALSE;
          seek(f,12);
          repeat
           blockread(f,buf^,8);
           fsize:=bswapl(ILBMChunk(buf^).size);
           if odd(fsize) then inc(fsize);
            With ILBMChunk(buf^) do
             if ((ChunkID[0]='C') and (ChunkID[1]='M') and
                 (ChunkId[2]='A') and (ChunkID[3]='P')) then
                  begin
                   Found:=TRUE;
                   blockread(f,pal,sizeof(pal)); {guaranteed this?}
                   for count:=0 to 255 do
                    with pal[count] do
                    begin
                     r:=r shr 2;
                     g:=g shr 2;
                     b:=b shr 2
                    end;
                  end
             else seek(f,filepos(f)+fsize);
            until (found) or (eof(f));
         end;
     BMP,COMPBMP:with Win3xBitmapHeader(buf^) do
         begin
          seek(f,size+sizeof(WinBMPFileHeader));
          for count:=0 to 255 do
           with pal[count] do
           begin
            blockread(f,plte,4);
            r:=plte.red shr 2;
            g:=plte.green shr 2;
            b:=plte.blue shr 2;
           end;
         end;
    end;
    system.freemem(buf,1024)
   end;
 close(f)
end;

procedure TGfx8.VertLine; external;

constructor TScreen8.init;
var MR,VR:word;
begin
 inherited init;
 _VESASelector:=AllocSelector(0);
 _CodeSelector:=0;
 _LFB:=FALSE;
 MR:=word(GlobalDosAlloc(sizeof(TMode_Rec)));
 VR:=word(GlobalDosAlloc(sizeof(TVesa_Rec)));
 ModeRec:=ptr(MR,0);
 VesaRec:=ptr(VR,0);
 _CurWriteBank:=65535;
 _CurReadBank:=65535;
 _InGraphics:=FALSE;
end;

procedure TScreen8.Clear; external;

procedure TScreen8.CopyBuf;
var count,xCount,yCount:integer;
begin
 if _LFB then inherited CopyBuf(DestX,DestY,SrcObject,x1,y1,x2,y2) else
  begin
   if SrcObject._data=0 then exit;
   if _Data=0 then exit;
   if SrcObject._data=Sega000 then exit; {do special case later}

   xCount:=(x2-x1+1);
   yCount:=(y2-y1+1);

   if (DestX+xCount>GetMaxX+1) then xCount:=GetMaxX-DestX+1;

   if (DestY+yCount>GetMaxY+1) then yCount:=GetMaxY-DestY+1;

   if (DestX<0) then
    begin
     inc(xcount,DestX);
     dec(x1,DestX);
     DestX:=0;
   end;

   if (DestY<0) then
    begin
     inc(yCount,DestY);
     dec(y1,DestY);
     DestY:=0;
    end;

   if (DestX+xCount>GetMaxX+1) or (DestX+xCount<0) or (DestY+yCount>GetMaxY+1) or
      (DestY+yCount<0) then exit;

    With SrcObject do
     for count:=0 to yCount-1 do
      CopyLineToScreen(ptr48(_data,_InternalOffset+LineOffsets[y1+count]+x1),
                       DestX,DestY+count,xCount);
 end;
end;

procedure Tscreen8.CopyLineFromScreen; external;
procedure TScreen8.CopyLineToScreen; external;

procedure TScreen8.Flip;
var count,xCount,yCount:word;
begin
 if SrcObject._data=0 then exit;
 if SrcObject._data=SegA000 then exit; {This would be flipping ourselves}
 if _LFB then inherited Flip(SrcObject) else
  begin
   xCount:=SrcObject.GetMaxX+1;
   if xCount>GetMaxX+1 then xCount:=GetMaxX+1;
   yCount:=SrcObject.GetMaxY;
   if yCount>GetMaxY then yCount:=GetMaxY;
   With SrcObject Do
   for count:=0 to yCount do
    CopyLineToScreen(ptr48(_data,_InternalOffset+LineOffsets[count]),
                     0,count,xCount);
  end;
end;

procedure TScreen8.GetImage; external;

function TScreen8.GetModeInfo;
begin
 with RMRegs do
  begin
   ax:=$4f01;
   cx:=mode;
   di:=0;
   es:=GetSegBaseAddr(seg(ModeRec^)) shr 4;
   sp:=0; {Clear stack so DPMI handles it}
   ss:=0  {"}
  end;
 RealModeInt($10,RMRegs);
 GetModeInfo:=RMRegs.ax;
end;

procedure TScreen8.getpal; assembler;
asm
        push    ds
        lds     di,self
        lea     si,TGfx8.ds:[di].pal
        mov     dx,03c7h
        xor     al,al
        out     dx,al
        add     dx,2
        mov     cx,768
@@lop:  in      al,dx
        mov     ds:[si],al
        inc     si
        dec     cx
        jnz     @@lop
        pop     ds
end;

procedure TScreen8.makepal; assembler;
asm
        push    ds
        lds     di,self
        lea     si,TGfx8.ds:[di].pal
        mov     dx,03c8h
        xor     al,al
        out     dx,al
        inc     dx
        mov     cx,768
@@lop:  mov     al,ds:[si]
        inc     si
        out     dx,al
        dec     cx
        jnz     @@lop
        pop     ds
end;

procedure TScreen8.HorizLine; external;
function TScreen8.GetPixel; external;

function TScreen8.IsModeAvailable(modeNum:word):boolean;
begin
 IsModeAvailable:=(ModeNum=$13);
 if (modeNum>=$100) then
  begin
   GetModeInfo(ModeNum);
   IsModeAvailable:=(ModeRec^.ModeAttributes and 1=1); {yeah, right}
  end;
end;

function TScreen8.IsVesaInstalled:Word;
begin
 with VesaRec^ do
  begin
   VBESignature[0]:='V';
   VBESignature[1]:='B';
   VBESignature[2]:='E';
   VBESignature[3]:='2';
  end;

 with RMRegs do
  begin
   ax:=$4f00;
   di:=0;
   es:=GetSegBaseAddr(seg(VESARec^)) shr 4;
   ss:=0; {Clear stack so DPMI handles it}
   sp:=0  {"}
  end;
 RealModeInt($10,RMRegs);

 {Convert all real mode pointers to protected mode pointers}
 with VesaRec^ do
  begin
   VideoModePtr:=ConvertPtr(VideoModePtr);
   OEMStringPtr:=ConvertPtr(OEMStringPtr);
   OEMVendorNamePtr:=ConvertPtr(OEMVendorNamePtr);
   OEMProductRevPtr:=ConvertPtr(OEMProductRevPtr);
  end;
 IsVesaInstalled:=RMRegs.ax
end;

procedure TScreen8.PutBlit; external;
procedure TScreen8.PutImage; external;
procedure TScreen8.PutPixel; external;
procedure TScreen8.RawLine; external;
procedure TScreen8.ScaleBuf; external;

procedure dummy; far; assembler;
asm end;

procedure OldSwitchbank; far; assembler;
asm
        push    ax
        mov     ax,4f05h
        int     10h
        pop     ax
end;

function TScreen8.setmode(mode:word):boolean;
begin
 SetMode:=FALSE; {presume failure}
 if IsVesaInstalled<>VESA_ok then exit;
 if not IsModeAvailable(mode) then exit;
 setupMode(mode);
 SetMode:=TRUE
end;

procedure TScreen8.SetRGBPalette;
begin
asm
        and     red,63
        and     green,63
        and     blue,63
        mov     al,colour
        mov     dx,03c8h
        out     dx,al
        inc     dx
        mov     al,red
        out     dx,al
        mov     al,green
        out     dx,al
        mov     al,blue
        out     dx,al
end;
 inherited setrgbpalette(colour,red,green,blue)
end;

procedure TScreen8.SetupMode;
var count,GranInc,Rights:word;
    PhysPtr,Size:longint;
    s:string[35];
begin
 {The IsVesaInstalled() call is possibly redundant, but necessary.  If
  someone calls SetupMode() without first calling IsVesaInstalled it will
  blow up.  Because SetMode checks the result of IsVesaInstalled, it is
  okay to throw away the result of it here.  Either they wanted error
  checking (in which case they called SetMode) or they don't want/need
  error checking and called this routine directly.}
 IsVesaInstalled;

 _Data:=Sega000;
 CurDataState:=ALIASING;
 _InGraphics:=TRUE;
 _LFB:=FALSE;
 if mode=$13 then
  begin
   InitVesamode(mode);
   asm
        mov     dx,03c2h
        mov     al,0e3h   {Init Mode-D}
        out     dx,al
   end;
   @SwitchBank:=@Dummy;
   xRes:=320;
   yRes:=200;
   GetMaxX:=Xres-1;
   GetMaxY:=YRes-1;
   LineOffsets[0]:=0;
   for count:=1 to high(LineOffsets) do
    LineOffsets[count]:=LineOffsets[count-1]+320;
   getpal;
   BPP:=8;
   exit
  end;

 if not(ForceBanked) and (VesaRec^.MajVersion>=2) then
  begin
   mode:=Mode or $4000; {attempt LFB}
   GetModeInfo(mode);
   with ModeRec^ do
    if physbaseptr=0 then Mode:=Mode and $3FFF  {No LFB, go banked}
     else
      begin
       size:=Xres*Longint(BytesPerLine);
       _LFB:=TRUE;
       PhysPtr:=PhysicalAddressMapping(PhysBasePtr,size);
       _data:=CreateDescriptor(PhysPtr,size);
       rights:=GetSegmentAccessRights(_Data);
       rights:=(rights and $FF70) or $0093;
       SetSegmentAccessRights(_data,rights)
      end
  end;

 GetModeInfo(mode);

 Xres:=ModeRec^.BytesPerLine;
 YRes:=ModeRec^.YRes;
 GetMaxX:=ModeRec^.Xres-1;
 GetMaxY:=YRes-1;

 LineOffsets[0]:=0;
 for count:=1 to high(LineOffsets) do
  LineOffsets[count]:=LineOffsets[count-1]+ModeRec^.BytesPerLine;

 GranInc:=1;
 if ModeRec^.Granularity<>0 then
  GranInc:=64 div ModeRec^.Granularity; {Adjust for page size}
  _BankVals[0]:=0;
 for count:=1 to high(_bankvals) do
  _BankVals[count]:=_BankVals[count-1]+GranInc;

 count:=0;
 with vesarec^ do
 while (count<32) and (pbytearray(OEMStringPtr)^[count]<>0) do
  begin
   s[count+1]:=upcase(char(pbytearray(OEMStringPtr)^[count]));
   inc(count);
  end;
 s[0]:=chr(count+1);
 if (pos('UNIVERSAL',s)>0) or (ForceInt10h) or (ModeRec^.BankSwitch=NIL)
  then @SwitchBank:=@OldSwitchBank
 else
  begin
   SetSelectorBase(_VESASelector,longint(dword(ModeRec^.BankSwitch).HiWord)*16);
   SetSelectorLimit(_VESASelector,65535);
   if _codeselector<>0 then FreeSelector(_CodeSelector);
   _CodeSelector:=AllocDStoCSAlias(_VesaSelector);
   @SwitchBank:=ptr(_CodeSelector,Dword(ModeRec^.BankSwitch).LoWord);
  end;
 If ModeRec^.WindowAFlags and 2=2 then
  _ReadWindow:=0
 else if ModeRec^.WindowBFlags and 2=2 then
  _ReadWindow:=1; {else there is no readable window!?}
 InitVESAMode(mode);
 BPP:=8;
 GetPal;
 _CurReadBank:=0;
 _CurWriteBank:=0;
end;

function TScreen8.UsingLFB;
begin
 UsingLFB:=_LFB
end;

procedure TScreen8.VertLine; external;

procedure TScreen8.WaitForRetrace; assembler; {Wait for Retrace}
asm
	mov     dx,03dah
@gn0:   in      al,dx
	test    al,8
	jnz     @gn0

@gn1:   in      al,dx
	test    al,8
	jz      @gn1
end;

destructor TScreen8.done;
begin
 inherited done;
 if _InGraphics then asm mov ax,3; int 10h end;
 if _VESASelector<>0 then FreeSelector(_VESASelector);
 if _CodeSelector<>0 then FreeSelector(_CodeSelector);
 if VesaRec<>NIL then globaldosfree(seg(VesaRec^));
 if ModeRec<>NIL then globaldosfree(seg(ModeRec^));
end;

begin

end.